/* This file is generated by python scripts. Don't edit this file directly.
 */

#ifndef CHEWING_ENUM_H
#define CHEWING_ENUM_H

namespace pinyin{

/**
 * @brief enums of chewing initial element.
 */

enum ChewingInitial
{
CHEWING_ZERO_INITIAL = 0,
CHEWING_B = 1,
CHEWING_C = 2,
CHEWING_CH = 3,
CHEWING_D = 4,
CHEWING_F = 5,
CHEWING_H = 6,
CHEWING_G = 7,
CHEWING_K = 8,
CHEWING_J = 9,
CHEWING_M = 10,
CHEWING_N = 11,
CHEWING_L = 12,
CHEWING_R = 13,
CHEWING_P = 14,
CHEWING_Q = 15,
CHEWING_S = 16,
CHEWING_SH = 17,
CHEWING_T = 18,
PINYIN_W = 19,
CHEWING_X = 20,
PINYIN_Y = 21,
CHEWING_Z = 22,
CHEWING_ZH = 23,
CHEWING_LAST_INITIAL = CHEWING_ZH,
CHEWING_NUMBER_OF_INITIALS = CHEWING_LAST_INITIAL + 1
};


/**
 * @brief enums of chewing middle element.
 */

enum ChewingMiddle
{
CHEWING_ZERO_MIDDLE = 0,
CHEWING_I = 1,
CHEWING_U = 2,
CHEWING_V = 3,
CHEWING_LAST_MIDDLE = CHEWING_V,
CHEWING_NUMBER_OF_MIDDLES = CHEWING_LAST_MIDDLE + 1
};


/**
 * @brief enums of chewing final element.
 */
enum ChewingFinal
{
CHEWING_ZERO_FINAL = 0,
CHEWING_A = 1,
CHEWING_AI = 2,
CHEWING_AN = 3,
CHEWING_ANG = 4,
CHEWING_AO = 5,
CHEWING_E = 6,
INVALID_EA = 7,
CHEWING_EI = 8,
CHEWING_EN = 9,
CHEWING_ENG = 10,
CHEWING_ER = 11,
CHEWING_NG = 12,
CHEWING_O = 13,
PINYIN_ONG = 14,
CHEWING_OU = 15,
PINYIN_IN = 16,
PINYIN_ING = 17,
CHEWING_LAST_FINAL = PINYIN_ING,
CHEWING_NUMBER_OF_FINALS = CHEWING_LAST_FINAL + 1
};


/**
 * @brief enums of chewing tone element.
 */
enum ChewingTone
{
CHEWING_ZERO_TONE = 0,
CHEWING_1 = 1,
CHEWING_2 = 2,
CHEWING_3 = 3,
CHEWING_4 = 4,
CHEWING_5 = 5,
CHEWING_LAST_TONE = CHEWING_5,
CHEWING_NUMBER_OF_TONES = CHEWING_LAST_TONE + 1
};

};

#endif
