/* This file is generated by python scripts. Don't edit this file directly.
 */

#ifndef PINYIN_PARSER_TABLE_H
#define PINYIN_PARSER_TABLE_H

namespace pinyin{

const pinyin_index_item_t pinyin_index[] = {
{"a", IS_CHEWING|IS_PINYIN, 1},
{"agn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 4},
{"ai", IS_CHEWING|IS_PINYIN, 2},
{"amg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 4},
{"an", IS_CHEWING|IS_PINYIN, 3},
{"ang", IS_CHEWING|IS_PINYIN, 4},
{"ao", IS_CHEWING|IS_PINYIN, 5},
{"b", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 6},
{"ba", IS_CHEWING|IS_PINYIN, 7},
{"bagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 10},
{"bai", IS_CHEWING|IS_PINYIN, 8},
{"bamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 10},
{"ban", IS_CHEWING|IS_PINYIN, 9},
{"bang", IS_CHEWING|IS_PINYIN, 10},
{"bao", IS_CHEWING|IS_PINYIN, 11},
{"begn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 14},
{"bei", IS_CHEWING|IS_PINYIN, 12},
{"bemg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 14},
{"ben", IS_CHEWING|IS_PINYIN, 13},
{"beng", IS_CHEWING|IS_PINYIN, 14},
{"bi", IS_CHEWING|IS_PINYIN, 15},
{"bian", IS_CHEWING|IS_PINYIN, 16},
{"biao", IS_CHEWING|IS_PINYIN, 17},
{"bie", IS_CHEWING|IS_PINYIN, 18},
{"bign", IS_PINYIN|PINYIN_CORRECT_GN_NG, 20},
{"bimg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 20},
{"bin", IS_CHEWING|IS_PINYIN, 19},
{"bing", IS_CHEWING|IS_PINYIN, 20},
{"bo", IS_CHEWING|IS_PINYIN, 21},
{"bu", IS_CHEWING|IS_PINYIN, 22},
{"c", IS_PINYIN|PINYIN_INCOMPLETE, 23},
{"ca", IS_CHEWING|IS_PINYIN, 24},
{"cagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 27},
{"cai", IS_CHEWING|IS_PINYIN, 25},
{"camg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 27},
{"can", IS_CHEWING|IS_PINYIN, 26},
{"cang", IS_CHEWING|IS_PINYIN, 27},
{"cao", IS_CHEWING|IS_PINYIN, 28},
{"ce", IS_CHEWING|IS_PINYIN, 29},
{"cegn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 31},
{"cemg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 31},
{"cen", IS_CHEWING|IS_PINYIN, 30},
{"ceng", IS_CHEWING|IS_PINYIN, 31},
{"ch", IS_PINYIN|PINYIN_INCOMPLETE, 32},
{"cha", IS_CHEWING|IS_PINYIN, 33},
{"chagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 36},
{"chai", IS_CHEWING|IS_PINYIN, 34},
{"chamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 36},
{"chan", IS_CHEWING|IS_PINYIN, 35},
{"chang", IS_CHEWING|IS_PINYIN, 36},
{"chao", IS_CHEWING|IS_PINYIN, 37},
{"che", IS_CHEWING|IS_PINYIN, 38},
{"chegn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 40},
{"chemg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 40},
{"chen", IS_CHEWING|IS_PINYIN, 39},
{"cheng", IS_CHEWING|IS_PINYIN, 40},
{"chi", IS_CHEWING|IS_PINYIN, 41},
{"chogn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 42},
{"chomg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 42},
{"chon", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 42},
{"chong", IS_CHEWING|IS_PINYIN, 42},
{"chou", IS_CHEWING|IS_PINYIN, 43},
{"chu", IS_CHEWING|IS_PINYIN, 44},
{"chuagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 48},
{"chuai", IS_CHEWING|IS_PINYIN, 46},
{"chuamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 48},
{"chuan", IS_CHEWING|IS_PINYIN, 47},
{"chuang", IS_CHEWING|IS_PINYIN, 48},
{"chuei", IS_PINYIN|PINYIN_CORRECT_UEI_UI, 49},
{"chuen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 50},
{"chui", IS_CHEWING|IS_PINYIN, 49},
{"chun", IS_CHEWING|IS_PINYIN, 50},
{"chuo", IS_CHEWING|IS_PINYIN, 51},
{"ci", IS_CHEWING|IS_PINYIN, 52},
{"cogn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 53},
{"comg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 53},
{"con", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 53},
{"cong", IS_CHEWING|IS_PINYIN, 53},
{"cou", IS_CHEWING|IS_PINYIN, 54},
{"cu", IS_CHEWING|IS_PINYIN, 55},
{"cuan", IS_CHEWING|IS_PINYIN, 56},
{"cuei", IS_PINYIN|PINYIN_CORRECT_UEI_UI, 57},
{"cuen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 58},
{"cui", IS_CHEWING|IS_PINYIN, 57},
{"cun", IS_CHEWING|IS_PINYIN, 58},
{"cuo", IS_CHEWING|IS_PINYIN, 59},
{"d", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 60},
{"da", IS_CHEWING|IS_PINYIN, 61},
{"dagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 64},
{"dai", IS_CHEWING|IS_PINYIN, 62},
{"damg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 64},
{"dan", IS_CHEWING|IS_PINYIN, 63},
{"dang", IS_CHEWING|IS_PINYIN, 64},
{"dao", IS_CHEWING|IS_PINYIN, 65},
{"de", IS_CHEWING|IS_PINYIN, 66},
{"degn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 69},
{"dei", IS_CHEWING|IS_PINYIN, 67},
{"demg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 69},
{"deng", IS_CHEWING|IS_PINYIN, 69},
{"di", IS_CHEWING|IS_PINYIN, 70},
{"dia", IS_CHEWING|IS_PINYIN, 71},
{"dian", IS_CHEWING|IS_PINYIN, 72},
{"diao", IS_CHEWING|IS_PINYIN, 73},
{"die", IS_CHEWING|IS_PINYIN, 74},
{"dign", IS_PINYIN|PINYIN_CORRECT_GN_NG, 76},
{"dimg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 76},
{"ding", IS_CHEWING|IS_PINYIN, 76},
{"diou", IS_PINYIN|PINYIN_CORRECT_IOU_IU, 77},
{"diu", IS_CHEWING|IS_PINYIN, 77},
{"dogn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 78},
{"domg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 78},
{"don", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 78},
{"dong", IS_CHEWING|IS_PINYIN, 78},
{"dou", IS_CHEWING|IS_PINYIN, 79},
{"du", IS_CHEWING|IS_PINYIN, 80},
{"duan", IS_CHEWING|IS_PINYIN, 81},
{"duei", IS_PINYIN|PINYIN_CORRECT_UEI_UI, 82},
{"duen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 83},
{"dui", IS_CHEWING|IS_PINYIN, 82},
{"dun", IS_CHEWING|IS_PINYIN, 83},
{"duo", IS_CHEWING|IS_PINYIN, 84},
{"e", IS_CHEWING|IS_PINYIN, 85},
{"egn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 88},
{"ei", IS_CHEWING|IS_PINYIN, 86},
{"emg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 88},
{"en", IS_CHEWING|IS_PINYIN, 87},
{"er", IS_CHEWING|IS_PINYIN, 89},
{"f", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 90},
{"fa", IS_CHEWING|IS_PINYIN, 91},
{"fagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 93},
{"famg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 93},
{"fan", IS_CHEWING|IS_PINYIN, 92},
{"fang", IS_CHEWING|IS_PINYIN, 93},
{"fegn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 97},
{"fei", IS_CHEWING|IS_PINYIN, 95},
{"femg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 97},
{"fen", IS_CHEWING|IS_PINYIN, 96},
{"feng", IS_CHEWING|IS_PINYIN, 97},
{"fo", IS_CHEWING|IS_PINYIN, 98},
{"fou", IS_CHEWING|IS_PINYIN, 99},
{"fu", IS_CHEWING|IS_PINYIN, 100},
{"g", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 101},
{"ga", IS_CHEWING|IS_PINYIN, 102},
{"gagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 105},
{"gai", IS_CHEWING|IS_PINYIN, 103},
{"gamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 105},
{"gan", IS_CHEWING|IS_PINYIN, 104},
{"gang", IS_CHEWING|IS_PINYIN, 105},
{"gao", IS_CHEWING|IS_PINYIN, 106},
{"ge", IS_CHEWING|IS_PINYIN, 107},
{"gegn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 110},
{"gei", IS_CHEWING|IS_PINYIN, 108},
{"gemg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 110},
{"gen", IS_CHEWING|IS_PINYIN, 109},
{"geng", IS_CHEWING|IS_PINYIN, 110},
{"gogn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 111},
{"gomg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 111},
{"gon", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 111},
{"gong", IS_CHEWING|IS_PINYIN, 111},
{"gou", IS_CHEWING|IS_PINYIN, 112},
{"gu", IS_CHEWING|IS_PINYIN, 113},
{"gua", IS_CHEWING|IS_PINYIN, 114},
{"guagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 117},
{"guai", IS_CHEWING|IS_PINYIN, 115},
{"guamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 117},
{"guan", IS_CHEWING|IS_PINYIN, 116},
{"guang", IS_CHEWING|IS_PINYIN, 117},
{"guei", IS_PINYIN|PINYIN_CORRECT_UEI_UI, 118},
{"guen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 119},
{"gui", IS_CHEWING|IS_PINYIN, 118},
{"gun", IS_CHEWING|IS_PINYIN, 119},
{"guo", IS_CHEWING|IS_PINYIN, 120},
{"h", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 121},
{"ha", IS_CHEWING|IS_PINYIN, 122},
{"hagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 125},
{"hai", IS_CHEWING|IS_PINYIN, 123},
{"hamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 125},
{"han", IS_CHEWING|IS_PINYIN, 124},
{"hang", IS_CHEWING|IS_PINYIN, 125},
{"hao", IS_CHEWING|IS_PINYIN, 126},
{"he", IS_CHEWING|IS_PINYIN, 127},
{"hegn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 130},
{"hei", IS_CHEWING|IS_PINYIN, 128},
{"hemg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 130},
{"hen", IS_CHEWING|IS_PINYIN, 129},
{"heng", IS_CHEWING|IS_PINYIN, 130},
{"hogn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 131},
{"homg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 131},
{"hon", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 131},
{"hong", IS_CHEWING|IS_PINYIN, 131},
{"hou", IS_CHEWING|IS_PINYIN, 132},
{"hu", IS_CHEWING|IS_PINYIN, 133},
{"hua", IS_CHEWING|IS_PINYIN, 134},
{"huagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 137},
{"huai", IS_CHEWING|IS_PINYIN, 135},
{"huamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 137},
{"huan", IS_CHEWING|IS_PINYIN, 136},
{"huang", IS_CHEWING|IS_PINYIN, 137},
{"huei", IS_PINYIN|PINYIN_CORRECT_UEI_UI, 138},
{"huen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 139},
{"hui", IS_CHEWING|IS_PINYIN, 138},
{"hun", IS_CHEWING|IS_PINYIN, 139},
{"huo", IS_CHEWING|IS_PINYIN, 140},
{"j", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 141},
{"ji", IS_CHEWING|IS_PINYIN, 142},
{"jia", IS_CHEWING|IS_PINYIN, 143},
{"jiagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 145},
{"jiamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 145},
{"jian", IS_CHEWING|IS_PINYIN, 144},
{"jiang", IS_CHEWING|IS_PINYIN, 145},
{"jiao", IS_CHEWING|IS_PINYIN, 146},
{"jie", IS_CHEWING|IS_PINYIN, 147},
{"jign", IS_PINYIN|PINYIN_CORRECT_GN_NG, 149},
{"jimg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 149},
{"jin", IS_CHEWING|IS_PINYIN, 148},
{"jing", IS_CHEWING|IS_PINYIN, 149},
{"jiogn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 150},
{"jiomg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 150},
{"jion", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 150},
{"jiong", IS_CHEWING|IS_PINYIN, 150},
{"jiou", IS_PINYIN|PINYIN_CORRECT_IOU_IU, 151},
{"jiu", IS_CHEWING|IS_PINYIN, 151},
{"ju", IS_CHEWING|IS_PINYIN, 152},
{"juan", IS_CHEWING|IS_PINYIN, 153},
{"jue", IS_CHEWING|IS_PINYIN, 154},
{"juen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 155},
{"jun", IS_CHEWING|IS_PINYIN, 155},
{"jv", IS_PINYIN|PINYIN_CORRECT_V_U, 152},
{"jvan", IS_PINYIN|PINYIN_CORRECT_V_U, 153},
{"jve", IS_PINYIN|PINYIN_CORRECT_V_U, 154},
{"jvn", IS_PINYIN|PINYIN_CORRECT_V_U, 155},
{"k", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 156},
{"ka", IS_CHEWING|IS_PINYIN, 157},
{"kagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 160},
{"kai", IS_CHEWING|IS_PINYIN, 158},
{"kamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 160},
{"kan", IS_CHEWING|IS_PINYIN, 159},
{"kang", IS_CHEWING|IS_PINYIN, 160},
{"kao", IS_CHEWING|IS_PINYIN, 161},
{"ke", IS_CHEWING|IS_PINYIN, 162},
{"kegn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 165},
{"kemg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 165},
{"ken", IS_CHEWING|IS_PINYIN, 164},
{"keng", IS_CHEWING|IS_PINYIN, 165},
{"kogn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 166},
{"komg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 166},
{"kon", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 166},
{"kong", IS_CHEWING|IS_PINYIN, 166},
{"kou", IS_CHEWING|IS_PINYIN, 167},
{"ku", IS_CHEWING|IS_PINYIN, 168},
{"kua", IS_CHEWING|IS_PINYIN, 169},
{"kuagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 172},
{"kuai", IS_CHEWING|IS_PINYIN, 170},
{"kuamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 172},
{"kuan", IS_CHEWING|IS_PINYIN, 171},
{"kuang", IS_CHEWING|IS_PINYIN, 172},
{"kuei", IS_PINYIN|PINYIN_CORRECT_UEI_UI, 173},
{"kuen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 174},
{"kui", IS_CHEWING|IS_PINYIN, 173},
{"kun", IS_CHEWING|IS_PINYIN, 174},
{"kuo", IS_CHEWING|IS_PINYIN, 175},
{"l", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 176},
{"la", IS_CHEWING|IS_PINYIN, 177},
{"lagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 180},
{"lai", IS_CHEWING|IS_PINYIN, 178},
{"lamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 180},
{"lan", IS_CHEWING|IS_PINYIN, 179},
{"lang", IS_CHEWING|IS_PINYIN, 180},
{"lao", IS_CHEWING|IS_PINYIN, 181},
{"le", IS_CHEWING|IS_PINYIN, 182},
{"legn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 185},
{"lei", IS_CHEWING|IS_PINYIN, 183},
{"lemg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 185},
{"leng", IS_CHEWING|IS_PINYIN, 185},
{"li", IS_CHEWING|IS_PINYIN, 186},
{"lia", IS_CHEWING|IS_PINYIN, 187},
{"liagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 189},
{"liamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 189},
{"lian", IS_CHEWING|IS_PINYIN, 188},
{"liang", IS_CHEWING|IS_PINYIN, 189},
{"liao", IS_CHEWING|IS_PINYIN, 190},
{"lie", IS_CHEWING|IS_PINYIN, 191},
{"lign", IS_PINYIN|PINYIN_CORRECT_GN_NG, 193},
{"limg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 193},
{"lin", IS_CHEWING|IS_PINYIN, 192},
{"ling", IS_CHEWING|IS_PINYIN, 193},
{"liou", IS_PINYIN|PINYIN_CORRECT_IOU_IU, 194},
{"liu", IS_CHEWING|IS_PINYIN, 194},
{"lo", IS_CHEWING|IS_PINYIN, 195},
{"logn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 196},
{"lomg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 196},
{"lon", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 196},
{"long", IS_CHEWING|IS_PINYIN, 196},
{"lou", IS_CHEWING|IS_PINYIN, 197},
{"lu", IS_CHEWING|IS_PINYIN, 198},
{"luan", IS_CHEWING|IS_PINYIN, 199},
{"lue", IS_PINYIN|PINYIN_CORRECT_UE_VE, 203},
{"luen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 200},
{"lun", IS_CHEWING|IS_PINYIN, 200},
{"luo", IS_CHEWING|IS_PINYIN, 201},
{"lv", IS_CHEWING|IS_PINYIN, 202},
{"lve", IS_CHEWING|IS_PINYIN, 203},
{"m", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 204},
{"ma", IS_CHEWING|IS_PINYIN, 205},
{"magn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 208},
{"mai", IS_CHEWING|IS_PINYIN, 206},
{"mamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 208},
{"man", IS_CHEWING|IS_PINYIN, 207},
{"mang", IS_CHEWING|IS_PINYIN, 208},
{"mao", IS_CHEWING|IS_PINYIN, 209},
{"me", IS_CHEWING|IS_PINYIN, 210},
{"megn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 213},
{"mei", IS_CHEWING|IS_PINYIN, 211},
{"memg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 213},
{"men", IS_CHEWING|IS_PINYIN, 212},
{"meng", IS_CHEWING|IS_PINYIN, 213},
{"mi", IS_CHEWING|IS_PINYIN, 214},
{"mian", IS_CHEWING|IS_PINYIN, 215},
{"miao", IS_CHEWING|IS_PINYIN, 216},
{"mie", IS_CHEWING|IS_PINYIN, 217},
{"mign", IS_PINYIN|PINYIN_CORRECT_GN_NG, 219},
{"mimg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 219},
{"min", IS_CHEWING|IS_PINYIN, 218},
{"ming", IS_CHEWING|IS_PINYIN, 219},
{"miou", IS_PINYIN|PINYIN_CORRECT_IOU_IU, 220},
{"miu", IS_CHEWING|IS_PINYIN, 220},
{"mo", IS_CHEWING|IS_PINYIN, 221},
{"mou", IS_CHEWING|IS_PINYIN, 222},
{"mu", IS_CHEWING|IS_PINYIN, 223},
{"n", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 224},
{"na", IS_CHEWING|IS_PINYIN, 225},
{"nagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 228},
{"nai", IS_CHEWING|IS_PINYIN, 226},
{"namg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 228},
{"nan", IS_CHEWING|IS_PINYIN, 227},
{"nang", IS_CHEWING|IS_PINYIN, 228},
{"nao", IS_CHEWING|IS_PINYIN, 229},
{"ne", IS_CHEWING|IS_PINYIN, 230},
{"negn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 233},
{"nei", IS_CHEWING|IS_PINYIN, 231},
{"nemg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 233},
{"nen", IS_CHEWING|IS_PINYIN, 232},
{"neng", IS_CHEWING|IS_PINYIN, 233},
{"ng", IS_CHEWING|IS_PINYIN, 234},
{"ni", IS_CHEWING|IS_PINYIN, 235},
{"niagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 238},
{"niamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 238},
{"nian", IS_CHEWING|IS_PINYIN, 237},
{"niang", IS_CHEWING|IS_PINYIN, 238},
{"niao", IS_CHEWING|IS_PINYIN, 239},
{"nie", IS_CHEWING|IS_PINYIN, 240},
{"nign", IS_PINYIN|PINYIN_CORRECT_GN_NG, 242},
{"nimg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 242},
{"nin", IS_CHEWING|IS_PINYIN, 241},
{"ning", IS_CHEWING|IS_PINYIN, 242},
{"niou", IS_PINYIN|PINYIN_CORRECT_IOU_IU, 243},
{"niu", IS_CHEWING|IS_PINYIN, 243},
{"nogn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 244},
{"nomg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 244},
{"non", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 244},
{"nong", IS_CHEWING|IS_PINYIN, 244},
{"nou", IS_CHEWING|IS_PINYIN, 245},
{"nu", IS_CHEWING|IS_PINYIN, 246},
{"nuan", IS_CHEWING|IS_PINYIN, 247},
{"nue", IS_PINYIN|PINYIN_CORRECT_UE_VE, 251},
{"nuen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 248},
{"nuo", IS_CHEWING|IS_PINYIN, 249},
{"nv", IS_CHEWING|IS_PINYIN, 250},
{"nve", IS_CHEWING|IS_PINYIN, 251},
{"o", IS_CHEWING|IS_PINYIN, 252},
{"ou", IS_CHEWING|IS_PINYIN, 253},
{"p", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 254},
{"pa", IS_CHEWING|IS_PINYIN, 255},
{"pagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 258},
{"pai", IS_CHEWING|IS_PINYIN, 256},
{"pamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 258},
{"pan", IS_CHEWING|IS_PINYIN, 257},
{"pang", IS_CHEWING|IS_PINYIN, 258},
{"pao", IS_CHEWING|IS_PINYIN, 259},
{"pegn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 262},
{"pei", IS_CHEWING|IS_PINYIN, 260},
{"pemg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 262},
{"pen", IS_CHEWING|IS_PINYIN, 261},
{"peng", IS_CHEWING|IS_PINYIN, 262},
{"pi", IS_CHEWING|IS_PINYIN, 263},
{"pian", IS_CHEWING|IS_PINYIN, 264},
{"piao", IS_CHEWING|IS_PINYIN, 265},
{"pie", IS_CHEWING|IS_PINYIN, 266},
{"pign", IS_PINYIN|PINYIN_CORRECT_GN_NG, 268},
{"pimg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 268},
{"pin", IS_CHEWING|IS_PINYIN, 267},
{"ping", IS_CHEWING|IS_PINYIN, 268},
{"po", IS_CHEWING|IS_PINYIN, 269},
{"pou", IS_CHEWING|IS_PINYIN, 270},
{"pu", IS_CHEWING|IS_PINYIN, 271},
{"q", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 272},
{"qi", IS_CHEWING|IS_PINYIN, 273},
{"qia", IS_CHEWING|IS_PINYIN, 274},
{"qiagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 276},
{"qiamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 276},
{"qian", IS_CHEWING|IS_PINYIN, 275},
{"qiang", IS_CHEWING|IS_PINYIN, 276},
{"qiao", IS_CHEWING|IS_PINYIN, 277},
{"qie", IS_CHEWING|IS_PINYIN, 278},
{"qign", IS_PINYIN|PINYIN_CORRECT_GN_NG, 280},
{"qimg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 280},
{"qin", IS_CHEWING|IS_PINYIN, 279},
{"qing", IS_CHEWING|IS_PINYIN, 280},
{"qiogn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 281},
{"qiomg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 281},
{"qion", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 281},
{"qiong", IS_CHEWING|IS_PINYIN, 281},
{"qiou", IS_PINYIN|PINYIN_CORRECT_IOU_IU, 282},
{"qiu", IS_CHEWING|IS_PINYIN, 282},
{"qu", IS_CHEWING|IS_PINYIN, 283},
{"quan", IS_CHEWING|IS_PINYIN, 284},
{"que", IS_CHEWING|IS_PINYIN, 285},
{"quen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 286},
{"qun", IS_CHEWING|IS_PINYIN, 286},
{"qv", IS_PINYIN|PINYIN_CORRECT_V_U, 283},
{"qvan", IS_PINYIN|PINYIN_CORRECT_V_U, 284},
{"qve", IS_PINYIN|PINYIN_CORRECT_V_U, 285},
{"qvn", IS_PINYIN|PINYIN_CORRECT_V_U, 286},
{"r", IS_PINYIN|PINYIN_INCOMPLETE, 287},
{"ragn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 289},
{"ramg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 289},
{"ran", IS_CHEWING|IS_PINYIN, 288},
{"rang", IS_CHEWING|IS_PINYIN, 289},
{"rao", IS_CHEWING|IS_PINYIN, 290},
{"re", IS_CHEWING|IS_PINYIN, 291},
{"regn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 293},
{"remg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 293},
{"ren", IS_CHEWING|IS_PINYIN, 292},
{"reng", IS_CHEWING|IS_PINYIN, 293},
{"ri", IS_CHEWING|IS_PINYIN, 294},
{"rogn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 295},
{"romg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 295},
{"ron", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 295},
{"rong", IS_CHEWING|IS_PINYIN, 295},
{"rou", IS_CHEWING|IS_PINYIN, 296},
{"ru", IS_CHEWING|IS_PINYIN, 297},
{"ruan", IS_CHEWING|IS_PINYIN, 299},
{"ruei", IS_PINYIN|PINYIN_CORRECT_UEI_UI, 300},
{"ruen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 301},
{"rui", IS_CHEWING|IS_PINYIN, 300},
{"run", IS_CHEWING|IS_PINYIN, 301},
{"ruo", IS_CHEWING|IS_PINYIN, 302},
{"s", IS_PINYIN|PINYIN_INCOMPLETE, 303},
{"sa", IS_CHEWING|IS_PINYIN, 304},
{"sagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 307},
{"sai", IS_CHEWING|IS_PINYIN, 305},
{"samg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 307},
{"san", IS_CHEWING|IS_PINYIN, 306},
{"sang", IS_CHEWING|IS_PINYIN, 307},
{"sao", IS_CHEWING|IS_PINYIN, 308},
{"se", IS_CHEWING|IS_PINYIN, 309},
{"segn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 311},
{"semg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 311},
{"sen", IS_CHEWING|IS_PINYIN, 310},
{"seng", IS_CHEWING|IS_PINYIN, 311},
{"sh", IS_PINYIN|PINYIN_INCOMPLETE, 312},
{"sha", IS_CHEWING|IS_PINYIN, 313},
{"shagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 316},
{"shai", IS_CHEWING|IS_PINYIN, 314},
{"shamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 316},
{"shan", IS_CHEWING|IS_PINYIN, 315},
{"shang", IS_CHEWING|IS_PINYIN, 316},
{"shao", IS_CHEWING|IS_PINYIN, 317},
{"she", IS_CHEWING|IS_PINYIN, 318},
{"shegn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 321},
{"shei", IS_CHEWING|IS_PINYIN, 319},
{"shemg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 321},
{"shen", IS_CHEWING|IS_PINYIN, 320},
{"sheng", IS_CHEWING|IS_PINYIN, 321},
{"shi", IS_CHEWING|IS_PINYIN, 322},
{"shou", IS_CHEWING|IS_PINYIN, 323},
{"shu", IS_CHEWING|IS_PINYIN, 324},
{"shua", IS_CHEWING|IS_PINYIN, 325},
{"shuagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 328},
{"shuai", IS_CHEWING|IS_PINYIN, 326},
{"shuamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 328},
{"shuan", IS_CHEWING|IS_PINYIN, 327},
{"shuang", IS_CHEWING|IS_PINYIN, 328},
{"shuei", IS_PINYIN|PINYIN_CORRECT_UEI_UI, 329},
{"shuen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 330},
{"shui", IS_CHEWING|IS_PINYIN, 329},
{"shun", IS_CHEWING|IS_PINYIN, 330},
{"shuo", IS_CHEWING|IS_PINYIN, 331},
{"si", IS_CHEWING|IS_PINYIN, 332},
{"sogn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 333},
{"somg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 333},
{"son", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 333},
{"song", IS_CHEWING|IS_PINYIN, 333},
{"sou", IS_CHEWING|IS_PINYIN, 334},
{"su", IS_CHEWING|IS_PINYIN, 335},
{"suan", IS_CHEWING|IS_PINYIN, 336},
{"suei", IS_PINYIN|PINYIN_CORRECT_UEI_UI, 337},
{"suen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 338},
{"sui", IS_CHEWING|IS_PINYIN, 337},
{"sun", IS_CHEWING|IS_PINYIN, 338},
{"suo", IS_CHEWING|IS_PINYIN, 339},
{"t", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 340},
{"ta", IS_CHEWING|IS_PINYIN, 341},
{"tagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 344},
{"tai", IS_CHEWING|IS_PINYIN, 342},
{"tamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 344},
{"tan", IS_CHEWING|IS_PINYIN, 343},
{"tang", IS_CHEWING|IS_PINYIN, 344},
{"tao", IS_CHEWING|IS_PINYIN, 345},
{"te", IS_CHEWING|IS_PINYIN, 346},
{"tegn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 347},
{"temg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 347},
{"teng", IS_CHEWING|IS_PINYIN, 347},
{"ti", IS_CHEWING|IS_PINYIN, 348},
{"tian", IS_CHEWING|IS_PINYIN, 349},
{"tiao", IS_CHEWING|IS_PINYIN, 350},
{"tie", IS_CHEWING|IS_PINYIN, 351},
{"tign", IS_PINYIN|PINYIN_CORRECT_GN_NG, 352},
{"timg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 352},
{"ting", IS_CHEWING|IS_PINYIN, 352},
{"togn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 353},
{"tomg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 353},
{"ton", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 353},
{"tong", IS_CHEWING|IS_PINYIN, 353},
{"tou", IS_CHEWING|IS_PINYIN, 354},
{"tu", IS_CHEWING|IS_PINYIN, 355},
{"tuan", IS_CHEWING|IS_PINYIN, 356},
{"tuei", IS_PINYIN|PINYIN_CORRECT_UEI_UI, 357},
{"tuen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 358},
{"tui", IS_CHEWING|IS_PINYIN, 357},
{"tun", IS_CHEWING|IS_PINYIN, 358},
{"tuo", IS_CHEWING|IS_PINYIN, 359},
{"w", IS_PINYIN|PINYIN_INCOMPLETE, 360},
{"wa", IS_CHEWING|IS_PINYIN, 361},
{"wagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 364},
{"wai", IS_CHEWING|IS_PINYIN, 362},
{"wamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 364},
{"wan", IS_CHEWING|IS_PINYIN, 363},
{"wang", IS_CHEWING|IS_PINYIN, 364},
{"wegn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 367},
{"wei", IS_CHEWING|IS_PINYIN, 365},
{"wemg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 367},
{"wen", IS_CHEWING|IS_PINYIN, 366},
{"weng", IS_CHEWING|IS_PINYIN, 367},
{"wo", IS_CHEWING|IS_PINYIN, 368},
{"wu", IS_CHEWING|IS_PINYIN, 369},
{"x", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 370},
{"xi", IS_CHEWING|IS_PINYIN, 371},
{"xia", IS_CHEWING|IS_PINYIN, 372},
{"xiagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 374},
{"xiamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 374},
{"xian", IS_CHEWING|IS_PINYIN, 373},
{"xiang", IS_CHEWING|IS_PINYIN, 374},
{"xiao", IS_CHEWING|IS_PINYIN, 375},
{"xie", IS_CHEWING|IS_PINYIN, 376},
{"xign", IS_PINYIN|PINYIN_CORRECT_GN_NG, 378},
{"ximg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 378},
{"xin", IS_CHEWING|IS_PINYIN, 377},
{"xing", IS_CHEWING|IS_PINYIN, 378},
{"xiogn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 379},
{"xiomg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 379},
{"xion", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 379},
{"xiong", IS_CHEWING|IS_PINYIN, 379},
{"xiou", IS_PINYIN|PINYIN_CORRECT_IOU_IU, 380},
{"xiu", IS_CHEWING|IS_PINYIN, 380},
{"xu", IS_CHEWING|IS_PINYIN, 381},
{"xuan", IS_CHEWING|IS_PINYIN, 382},
{"xue", IS_CHEWING|IS_PINYIN, 383},
{"xuen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 384},
{"xun", IS_CHEWING|IS_PINYIN, 384},
{"xv", IS_PINYIN|PINYIN_CORRECT_V_U, 381},
{"xvan", IS_PINYIN|PINYIN_CORRECT_V_U, 382},
{"xve", IS_PINYIN|PINYIN_CORRECT_V_U, 383},
{"xvn", IS_PINYIN|PINYIN_CORRECT_V_U, 384},
{"y", IS_PINYIN|PINYIN_INCOMPLETE, 385},
{"ya", IS_CHEWING|IS_PINYIN, 386},
{"yagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 389},
{"yamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 389},
{"yan", IS_CHEWING|IS_PINYIN, 388},
{"yang", IS_CHEWING|IS_PINYIN, 389},
{"yao", IS_CHEWING|IS_PINYIN, 390},
{"ye", IS_CHEWING|IS_PINYIN, 391},
{"yi", IS_CHEWING|IS_PINYIN, 392},
{"yign", IS_PINYIN|PINYIN_CORRECT_GN_NG, 394},
{"yimg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 394},
{"yin", IS_CHEWING|IS_PINYIN, 393},
{"ying", IS_CHEWING|IS_PINYIN, 394},
{"yo", IS_CHEWING|IS_PINYIN, 395},
{"yogn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 396},
{"yomg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 396},
{"yon", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 396},
{"yong", IS_CHEWING|IS_PINYIN, 396},
{"you", IS_CHEWING|IS_PINYIN, 397},
{"yu", IS_CHEWING|IS_PINYIN, 398},
{"yuan", IS_CHEWING|IS_PINYIN, 399},
{"yue", IS_CHEWING|IS_PINYIN, 400},
{"yuen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 401},
{"yun", IS_CHEWING|IS_PINYIN, 401},
{"yv", IS_PINYIN|PINYIN_CORRECT_V_U, 398},
{"yvan", IS_PINYIN|PINYIN_CORRECT_V_U, 399},
{"yve", IS_PINYIN|PINYIN_CORRECT_V_U, 400},
{"yvn", IS_PINYIN|PINYIN_CORRECT_V_U, 401},
{"z", IS_PINYIN|PINYIN_INCOMPLETE, 402},
{"za", IS_CHEWING|IS_PINYIN, 403},
{"zagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 406},
{"zai", IS_CHEWING|IS_PINYIN, 404},
{"zamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 406},
{"zan", IS_CHEWING|IS_PINYIN, 405},
{"zang", IS_CHEWING|IS_PINYIN, 406},
{"zao", IS_CHEWING|IS_PINYIN, 407},
{"ze", IS_CHEWING|IS_PINYIN, 408},
{"zegn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 411},
{"zei", IS_CHEWING|IS_PINYIN, 409},
{"zemg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 411},
{"zen", IS_CHEWING|IS_PINYIN, 410},
{"zeng", IS_CHEWING|IS_PINYIN, 411},
{"zh", IS_PINYIN|PINYIN_INCOMPLETE, 412},
{"zha", IS_CHEWING|IS_PINYIN, 413},
{"zhagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 416},
{"zhai", IS_CHEWING|IS_PINYIN, 414},
{"zhamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 416},
{"zhan", IS_CHEWING|IS_PINYIN, 415},
{"zhang", IS_CHEWING|IS_PINYIN, 416},
{"zhao", IS_CHEWING|IS_PINYIN, 417},
{"zhe", IS_CHEWING|IS_PINYIN, 418},
{"zhegn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 421},
{"zhemg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 421},
{"zhen", IS_CHEWING|IS_PINYIN, 420},
{"zheng", IS_CHEWING|IS_PINYIN, 421},
{"zhi", IS_CHEWING|IS_PINYIN, 422},
{"zhogn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 423},
{"zhomg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 423},
{"zhon", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 423},
{"zhong", IS_CHEWING|IS_PINYIN, 423},
{"zhou", IS_CHEWING|IS_PINYIN, 424},
{"zhu", IS_CHEWING|IS_PINYIN, 425},
{"zhua", IS_CHEWING|IS_PINYIN, 426},
{"zhuagn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 429},
{"zhuai", IS_CHEWING|IS_PINYIN, 427},
{"zhuamg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 429},
{"zhuan", IS_CHEWING|IS_PINYIN, 428},
{"zhuang", IS_CHEWING|IS_PINYIN, 429},
{"zhuei", IS_PINYIN|PINYIN_CORRECT_UEI_UI, 430},
{"zhuen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 431},
{"zhui", IS_CHEWING|IS_PINYIN, 430},
{"zhun", IS_CHEWING|IS_PINYIN, 431},
{"zhuo", IS_CHEWING|IS_PINYIN, 432},
{"zi", IS_CHEWING|IS_PINYIN, 433},
{"zogn", IS_PINYIN|PINYIN_CORRECT_GN_NG, 434},
{"zomg", IS_PINYIN|PINYIN_CORRECT_MG_NG, 434},
{"zon", IS_PINYIN|PINYIN_CORRECT_ON_ONG, 434},
{"zong", IS_CHEWING|IS_PINYIN, 434},
{"zou", IS_CHEWING|IS_PINYIN, 435},
{"zu", IS_CHEWING|IS_PINYIN, 436},
{"zuan", IS_CHEWING|IS_PINYIN, 437},
{"zuei", IS_PINYIN|PINYIN_CORRECT_UEI_UI, 438},
{"zuen", IS_PINYIN|PINYIN_CORRECT_UEN_UN, 439},
{"zui", IS_CHEWING|IS_PINYIN, 438},
{"zun", IS_CHEWING|IS_PINYIN, 439},
{"zuo", IS_CHEWING|IS_PINYIN, 440}
};

const chewing_index_item_t chewing_index[] = {
{"ㄅ", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 6},
{"ㄅㄚ", IS_CHEWING|IS_PINYIN, 7},
{"ㄅㄛ", IS_CHEWING|IS_PINYIN, 21},
{"ㄅㄞ", IS_CHEWING|IS_PINYIN, 8},
{"ㄅㄟ", IS_CHEWING|IS_PINYIN, 12},
{"ㄅㄠ", IS_CHEWING|IS_PINYIN, 11},
{"ㄅㄢ", IS_CHEWING|IS_PINYIN, 9},
{"ㄅㄣ", IS_CHEWING|IS_PINYIN, 13},
{"ㄅㄤ", IS_CHEWING|IS_PINYIN, 10},
{"ㄅㄥ", IS_CHEWING|IS_PINYIN, 14},
{"ㄅㄧ", IS_CHEWING|IS_PINYIN, 15},
{"ㄅㄧㄝ", IS_CHEWING|IS_PINYIN, 18},
{"ㄅㄧㄠ", IS_CHEWING|IS_PINYIN, 17},
{"ㄅㄧㄢ", IS_CHEWING|IS_PINYIN, 16},
{"ㄅㄧㄣ", IS_CHEWING|IS_PINYIN, 19},
{"ㄅㄧㄥ", IS_CHEWING|IS_PINYIN, 20},
{"ㄅㄨ", IS_CHEWING|IS_PINYIN, 22},
{"ㄆ", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 254},
{"ㄆㄚ", IS_CHEWING|IS_PINYIN, 255},
{"ㄆㄛ", IS_CHEWING|IS_PINYIN, 269},
{"ㄆㄞ", IS_CHEWING|IS_PINYIN, 256},
{"ㄆㄟ", IS_CHEWING|IS_PINYIN, 260},
{"ㄆㄠ", IS_CHEWING|IS_PINYIN, 259},
{"ㄆㄡ", IS_CHEWING|IS_PINYIN, 270},
{"ㄆㄢ", IS_CHEWING|IS_PINYIN, 257},
{"ㄆㄣ", IS_CHEWING|IS_PINYIN, 261},
{"ㄆㄤ", IS_CHEWING|IS_PINYIN, 258},
{"ㄆㄥ", IS_CHEWING|IS_PINYIN, 262},
{"ㄆㄧ", IS_CHEWING|IS_PINYIN, 263},
{"ㄆㄧㄝ", IS_CHEWING|IS_PINYIN, 266},
{"ㄆㄧㄠ", IS_CHEWING|IS_PINYIN, 265},
{"ㄆㄧㄢ", IS_CHEWING|IS_PINYIN, 264},
{"ㄆㄧㄣ", IS_CHEWING|IS_PINYIN, 267},
{"ㄆㄧㄥ", IS_CHEWING|IS_PINYIN, 268},
{"ㄆㄨ", IS_CHEWING|IS_PINYIN, 271},
{"ㄇ", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 204},
{"ㄇㄚ", IS_CHEWING|IS_PINYIN, 205},
{"ㄇㄛ", IS_CHEWING|IS_PINYIN, 221},
{"ㄇㄜ", IS_CHEWING|IS_PINYIN, 210},
{"ㄇㄞ", IS_CHEWING|IS_PINYIN, 206},
{"ㄇㄟ", IS_CHEWING|IS_PINYIN, 211},
{"ㄇㄠ", IS_CHEWING|IS_PINYIN, 209},
{"ㄇㄡ", IS_CHEWING|IS_PINYIN, 222},
{"ㄇㄢ", IS_CHEWING|IS_PINYIN, 207},
{"ㄇㄣ", IS_CHEWING|IS_PINYIN, 212},
{"ㄇㄤ", IS_CHEWING|IS_PINYIN, 208},
{"ㄇㄥ", IS_CHEWING|IS_PINYIN, 213},
{"ㄇㄧ", IS_CHEWING|IS_PINYIN, 214},
{"ㄇㄧㄝ", IS_CHEWING|IS_PINYIN, 217},
{"ㄇㄧㄠ", IS_CHEWING|IS_PINYIN, 216},
{"ㄇㄧㄡ", IS_CHEWING|IS_PINYIN, 220},
{"ㄇㄧㄢ", IS_CHEWING|IS_PINYIN, 215},
{"ㄇㄧㄣ", IS_CHEWING|IS_PINYIN, 218},
{"ㄇㄧㄥ", IS_CHEWING|IS_PINYIN, 219},
{"ㄇㄨ", IS_CHEWING|IS_PINYIN, 223},
{"ㄈ", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 90},
{"ㄈㄚ", IS_CHEWING|IS_PINYIN, 91},
{"ㄈㄛ", IS_CHEWING|IS_PINYIN, 98},
{"ㄈㄜ", IS_CHEWING, 94},
{"ㄈㄟ", IS_CHEWING|IS_PINYIN, 95},
{"ㄈㄡ", IS_CHEWING|IS_PINYIN, 99},
{"ㄈㄢ", IS_CHEWING|IS_PINYIN, 92},
{"ㄈㄣ", IS_CHEWING|IS_PINYIN, 96},
{"ㄈㄤ", IS_CHEWING|IS_PINYIN, 93},
{"ㄈㄥ", IS_CHEWING|IS_PINYIN, 97},
{"ㄈㄨ", IS_CHEWING|IS_PINYIN, 100},
{"ㄉ", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 60},
{"ㄉㄚ", IS_CHEWING|IS_PINYIN, 61},
{"ㄉㄜ", IS_CHEWING|IS_PINYIN, 66},
{"ㄉㄞ", IS_CHEWING|IS_PINYIN, 62},
{"ㄉㄟ", IS_CHEWING|IS_PINYIN, 67},
{"ㄉㄠ", IS_CHEWING|IS_PINYIN, 65},
{"ㄉㄡ", IS_CHEWING|IS_PINYIN, 79},
{"ㄉㄢ", IS_CHEWING|IS_PINYIN, 63},
{"ㄉㄣ", IS_CHEWING, 68},
{"ㄉㄤ", IS_CHEWING|IS_PINYIN, 64},
{"ㄉㄥ", IS_CHEWING|IS_PINYIN, 69},
{"ㄉㄧ", IS_CHEWING|IS_PINYIN, 70},
{"ㄉㄧㄚ", IS_CHEWING|IS_PINYIN, 71},
{"ㄉㄧㄝ", IS_CHEWING|IS_PINYIN, 74},
{"ㄉㄧㄠ", IS_CHEWING|IS_PINYIN, 73},
{"ㄉㄧㄡ", IS_CHEWING|IS_PINYIN, 77},
{"ㄉㄧㄢ", IS_CHEWING|IS_PINYIN, 72},
{"ㄉㄧㄣ", IS_CHEWING, 75},
{"ㄉㄧㄥ", IS_CHEWING|IS_PINYIN, 76},
{"ㄉㄨ", IS_CHEWING|IS_PINYIN, 80},
{"ㄉㄨㄛ", IS_CHEWING|IS_PINYIN, 84},
{"ㄉㄨㄟ", IS_CHEWING|IS_PINYIN, 82},
{"ㄉㄨㄢ", IS_CHEWING|IS_PINYIN, 81},
{"ㄉㄨㄣ", IS_CHEWING|IS_PINYIN, 83},
{"ㄉㄨㄥ", IS_CHEWING|IS_PINYIN, 78},
{"ㄊ", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 340},
{"ㄊㄚ", IS_CHEWING|IS_PINYIN, 341},
{"ㄊㄜ", IS_CHEWING|IS_PINYIN, 346},
{"ㄊㄞ", IS_CHEWING|IS_PINYIN, 342},
{"ㄊㄠ", IS_CHEWING|IS_PINYIN, 345},
{"ㄊㄡ", IS_CHEWING|IS_PINYIN, 354},
{"ㄊㄢ", IS_CHEWING|IS_PINYIN, 343},
{"ㄊㄤ", IS_CHEWING|IS_PINYIN, 344},
{"ㄊㄥ", IS_CHEWING|IS_PINYIN, 347},
{"ㄊㄧ", IS_CHEWING|IS_PINYIN, 348},
{"ㄊㄧㄝ", IS_CHEWING|IS_PINYIN, 351},
{"ㄊㄧㄠ", IS_CHEWING|IS_PINYIN, 350},
{"ㄊㄧㄢ", IS_CHEWING|IS_PINYIN, 349},
{"ㄊㄧㄥ", IS_CHEWING|IS_PINYIN, 352},
{"ㄊㄨ", IS_CHEWING|IS_PINYIN, 355},
{"ㄊㄨㄛ", IS_CHEWING|IS_PINYIN, 359},
{"ㄊㄨㄟ", IS_CHEWING|IS_PINYIN, 357},
{"ㄊㄨㄢ", IS_CHEWING|IS_PINYIN, 356},
{"ㄊㄨㄣ", IS_CHEWING|IS_PINYIN, 358},
{"ㄊㄨㄥ", IS_CHEWING|IS_PINYIN, 353},
{"ㄋ", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 224},
{"ㄋㄚ", IS_CHEWING|IS_PINYIN, 225},
{"ㄋㄜ", IS_CHEWING|IS_PINYIN, 230},
{"ㄋㄞ", IS_CHEWING|IS_PINYIN, 226},
{"ㄋㄟ", IS_CHEWING|IS_PINYIN, 231},
{"ㄋㄠ", IS_CHEWING|IS_PINYIN, 229},
{"ㄋㄡ", IS_CHEWING|IS_PINYIN, 245},
{"ㄋㄢ", IS_CHEWING|IS_PINYIN, 227},
{"ㄋㄣ", IS_CHEWING|IS_PINYIN, 232},
{"ㄋㄤ", IS_CHEWING|IS_PINYIN, 228},
{"ㄋㄥ", IS_CHEWING|IS_PINYIN, 233},
{"ㄋㄧ", IS_CHEWING|IS_PINYIN, 235},
{"ㄋㄧㄚ", IS_CHEWING, 236},
{"ㄋㄧㄝ", IS_CHEWING|IS_PINYIN, 240},
{"ㄋㄧㄠ", IS_CHEWING|IS_PINYIN, 239},
{"ㄋㄧㄡ", IS_CHEWING|IS_PINYIN, 243},
{"ㄋㄧㄢ", IS_CHEWING|IS_PINYIN, 237},
{"ㄋㄧㄣ", IS_CHEWING|IS_PINYIN, 241},
{"ㄋㄧㄤ", IS_CHEWING|IS_PINYIN, 238},
{"ㄋㄧㄥ", IS_CHEWING|IS_PINYIN, 242},
{"ㄋㄨ", IS_CHEWING|IS_PINYIN, 246},
{"ㄋㄨㄛ", IS_CHEWING|IS_PINYIN, 249},
{"ㄋㄨㄢ", IS_CHEWING|IS_PINYIN, 247},
{"ㄋㄨㄣ", IS_CHEWING, 248},
{"ㄋㄨㄥ", IS_CHEWING|IS_PINYIN, 244},
{"ㄋㄩ", IS_CHEWING|IS_PINYIN, 250},
{"ㄋㄩㄝ", IS_CHEWING|IS_PINYIN, 251},
{"ㄌ", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 176},
{"ㄌㄚ", IS_CHEWING|IS_PINYIN, 177},
{"ㄌㄛ", IS_CHEWING|IS_PINYIN, 195},
{"ㄌㄜ", IS_CHEWING|IS_PINYIN, 182},
{"ㄌㄞ", IS_CHEWING|IS_PINYIN, 178},
{"ㄌㄟ", IS_CHEWING|IS_PINYIN, 183},
{"ㄌㄠ", IS_CHEWING|IS_PINYIN, 181},
{"ㄌㄡ", IS_CHEWING|IS_PINYIN, 197},
{"ㄌㄢ", IS_CHEWING|IS_PINYIN, 179},
{"ㄌㄣ", IS_CHEWING, 184},
{"ㄌㄤ", IS_CHEWING|IS_PINYIN, 180},
{"ㄌㄥ", IS_CHEWING|IS_PINYIN, 185},
{"ㄌㄧ", IS_CHEWING|IS_PINYIN, 186},
{"ㄌㄧㄚ", IS_CHEWING|IS_PINYIN, 187},
{"ㄌㄧㄝ", IS_CHEWING|IS_PINYIN, 191},
{"ㄌㄧㄠ", IS_CHEWING|IS_PINYIN, 190},
{"ㄌㄧㄡ", IS_CHEWING|IS_PINYIN, 194},
{"ㄌㄧㄢ", IS_CHEWING|IS_PINYIN, 188},
{"ㄌㄧㄣ", IS_CHEWING|IS_PINYIN, 192},
{"ㄌㄧㄤ", IS_CHEWING|IS_PINYIN, 189},
{"ㄌㄧㄥ", IS_CHEWING|IS_PINYIN, 193},
{"ㄌㄨ", IS_CHEWING|IS_PINYIN, 198},
{"ㄌㄨㄛ", IS_CHEWING|IS_PINYIN, 201},
{"ㄌㄨㄢ", IS_CHEWING|IS_PINYIN, 199},
{"ㄌㄨㄣ", IS_CHEWING|IS_PINYIN, 200},
{"ㄌㄨㄥ", IS_CHEWING|IS_PINYIN, 196},
{"ㄌㄩ", IS_CHEWING|IS_PINYIN, 202},
{"ㄌㄩㄝ", IS_CHEWING|IS_PINYIN, 203},
{"ㄍ", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 101},
{"ㄍㄚ", IS_CHEWING|IS_PINYIN, 102},
{"ㄍㄜ", IS_CHEWING|IS_PINYIN, 107},
{"ㄍㄞ", IS_CHEWING|IS_PINYIN, 103},
{"ㄍㄟ", IS_CHEWING|IS_PINYIN, 108},
{"ㄍㄠ", IS_CHEWING|IS_PINYIN, 106},
{"ㄍㄡ", IS_CHEWING|IS_PINYIN, 112},
{"ㄍㄢ", IS_CHEWING|IS_PINYIN, 104},
{"ㄍㄣ", IS_CHEWING|IS_PINYIN, 109},
{"ㄍㄤ", IS_CHEWING|IS_PINYIN, 105},
{"ㄍㄥ", IS_CHEWING|IS_PINYIN, 110},
{"ㄍㄨ", IS_CHEWING|IS_PINYIN, 113},
{"ㄍㄨㄚ", IS_CHEWING|IS_PINYIN, 114},
{"ㄍㄨㄛ", IS_CHEWING|IS_PINYIN, 120},
{"ㄍㄨㄞ", IS_CHEWING|IS_PINYIN, 115},
{"ㄍㄨㄟ", IS_CHEWING|IS_PINYIN, 118},
{"ㄍㄨㄢ", IS_CHEWING|IS_PINYIN, 116},
{"ㄍㄨㄣ", IS_CHEWING|IS_PINYIN, 119},
{"ㄍㄨㄤ", IS_CHEWING|IS_PINYIN, 117},
{"ㄍㄨㄥ", IS_CHEWING|IS_PINYIN, 111},
{"ㄎ", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 156},
{"ㄎㄚ", IS_CHEWING|IS_PINYIN, 157},
{"ㄎㄜ", IS_CHEWING|IS_PINYIN, 162},
{"ㄎㄞ", IS_CHEWING|IS_PINYIN, 158},
{"ㄎㄟ", IS_CHEWING, 163},
{"ㄎㄠ", IS_CHEWING|IS_PINYIN, 161},
{"ㄎㄡ", IS_CHEWING|IS_PINYIN, 167},
{"ㄎㄢ", IS_CHEWING|IS_PINYIN, 159},
{"ㄎㄣ", IS_CHEWING|IS_PINYIN, 164},
{"ㄎㄤ", IS_CHEWING|IS_PINYIN, 160},
{"ㄎㄥ", IS_CHEWING|IS_PINYIN, 165},
{"ㄎㄨ", IS_CHEWING|IS_PINYIN, 168},
{"ㄎㄨㄚ", IS_CHEWING|IS_PINYIN, 169},
{"ㄎㄨㄛ", IS_CHEWING|IS_PINYIN, 175},
{"ㄎㄨㄞ", IS_CHEWING|IS_PINYIN, 170},
{"ㄎㄨㄟ", IS_CHEWING|IS_PINYIN, 173},
{"ㄎㄨㄢ", IS_CHEWING|IS_PINYIN, 171},
{"ㄎㄨㄣ", IS_CHEWING|IS_PINYIN, 174},
{"ㄎㄨㄤ", IS_CHEWING|IS_PINYIN, 172},
{"ㄎㄨㄥ", IS_CHEWING|IS_PINYIN, 166},
{"ㄏ", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 121},
{"ㄏㄚ", IS_CHEWING|IS_PINYIN, 122},
{"ㄏㄜ", IS_CHEWING|IS_PINYIN, 127},
{"ㄏㄞ", IS_CHEWING|IS_PINYIN, 123},
{"ㄏㄟ", IS_CHEWING|IS_PINYIN, 128},
{"ㄏㄠ", IS_CHEWING|IS_PINYIN, 126},
{"ㄏㄡ", IS_CHEWING|IS_PINYIN, 132},
{"ㄏㄢ", IS_CHEWING|IS_PINYIN, 124},
{"ㄏㄣ", IS_CHEWING|IS_PINYIN, 129},
{"ㄏㄤ", IS_CHEWING|IS_PINYIN, 125},
{"ㄏㄥ", IS_CHEWING|IS_PINYIN, 130},
{"ㄏㄨ", IS_CHEWING|IS_PINYIN, 133},
{"ㄏㄨㄚ", IS_CHEWING|IS_PINYIN, 134},
{"ㄏㄨㄛ", IS_CHEWING|IS_PINYIN, 140},
{"ㄏㄨㄞ", IS_CHEWING|IS_PINYIN, 135},
{"ㄏㄨㄟ", IS_CHEWING|IS_PINYIN, 138},
{"ㄏㄨㄢ", IS_CHEWING|IS_PINYIN, 136},
{"ㄏㄨㄣ", IS_CHEWING|IS_PINYIN, 139},
{"ㄏㄨㄤ", IS_CHEWING|IS_PINYIN, 137},
{"ㄏㄨㄥ", IS_CHEWING|IS_PINYIN, 131},
{"ㄐ", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 141},
{"ㄐㄧ", IS_CHEWING|IS_PINYIN, 142},
{"ㄐㄧㄚ", IS_CHEWING|IS_PINYIN, 143},
{"ㄐㄧㄝ", IS_CHEWING|IS_PINYIN, 147},
{"ㄐㄧㄠ", IS_CHEWING|IS_PINYIN, 146},
{"ㄐㄧㄡ", IS_CHEWING|IS_PINYIN, 151},
{"ㄐㄧㄢ", IS_CHEWING|IS_PINYIN, 144},
{"ㄐㄧㄣ", IS_CHEWING|IS_PINYIN, 148},
{"ㄐㄧㄤ", IS_CHEWING|IS_PINYIN, 145},
{"ㄐㄧㄥ", IS_CHEWING|IS_PINYIN, 149},
{"ㄐㄩ", IS_CHEWING|IS_PINYIN, 152},
{"ㄐㄩㄝ", IS_CHEWING|IS_PINYIN, 154},
{"ㄐㄩㄢ", IS_CHEWING|IS_PINYIN, 153},
{"ㄐㄩㄣ", IS_CHEWING|IS_PINYIN, 155},
{"ㄐㄩㄥ", IS_CHEWING|IS_PINYIN, 150},
{"ㄑ", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 272},
{"ㄑㄧ", IS_CHEWING|IS_PINYIN, 273},
{"ㄑㄧㄚ", IS_CHEWING|IS_PINYIN, 274},
{"ㄑㄧㄝ", IS_CHEWING|IS_PINYIN, 278},
{"ㄑㄧㄠ", IS_CHEWING|IS_PINYIN, 277},
{"ㄑㄧㄡ", IS_CHEWING|IS_PINYIN, 282},
{"ㄑㄧㄢ", IS_CHEWING|IS_PINYIN, 275},
{"ㄑㄧㄣ", IS_CHEWING|IS_PINYIN, 279},
{"ㄑㄧㄤ", IS_CHEWING|IS_PINYIN, 276},
{"ㄑㄧㄥ", IS_CHEWING|IS_PINYIN, 280},
{"ㄑㄩ", IS_CHEWING|IS_PINYIN, 283},
{"ㄑㄩㄝ", IS_CHEWING|IS_PINYIN, 285},
{"ㄑㄩㄢ", IS_CHEWING|IS_PINYIN, 284},
{"ㄑㄩㄣ", IS_CHEWING|IS_PINYIN, 286},
{"ㄑㄩㄥ", IS_CHEWING|IS_PINYIN, 281},
{"ㄒ", IS_CHEWING|IS_PINYIN|PINYIN_INCOMPLETE|CHEWING_INCOMPLETE, 370},
{"ㄒㄧ", IS_CHEWING|IS_PINYIN, 371},
{"ㄒㄧㄚ", IS_CHEWING|IS_PINYIN, 372},
{"ㄒㄧㄝ", IS_CHEWING|IS_PINYIN, 376},
{"ㄒㄧㄠ", IS_CHEWING|IS_PINYIN, 375},
{"ㄒㄧㄡ", IS_CHEWING|IS_PINYIN, 380},
{"ㄒㄧㄢ", IS_CHEWING|IS_PINYIN, 373},
{"ㄒㄧㄣ", IS_CHEWING|IS_PINYIN, 377},
{"ㄒㄧㄤ", IS_CHEWING|IS_PINYIN, 374},
{"ㄒㄧㄥ", IS_CHEWING|IS_PINYIN, 378},
{"ㄒㄩ", IS_CHEWING|IS_PINYIN, 381},
{"ㄒㄩㄝ", IS_CHEWING|IS_PINYIN, 383},
{"ㄒㄩㄢ", IS_CHEWING|IS_PINYIN, 382},
{"ㄒㄩㄣ", IS_CHEWING|IS_PINYIN, 384},
{"ㄒㄩㄥ", IS_CHEWING|IS_PINYIN, 379},
{"ㄓ", IS_CHEWING|IS_PINYIN, 422},
{"ㄓㄚ", IS_CHEWING|IS_PINYIN, 413},
{"ㄓㄜ", IS_CHEWING|IS_PINYIN, 418},
{"ㄓㄞ", IS_CHEWING|IS_PINYIN, 414},
{"ㄓㄟ", IS_CHEWING, 419},
{"ㄓㄠ", IS_CHEWING|IS_PINYIN, 417},
{"ㄓㄡ", IS_CHEWING|IS_PINYIN, 424},
{"ㄓㄢ", IS_CHEWING|IS_PINYIN, 415},
{"ㄓㄣ", IS_CHEWING|IS_PINYIN, 420},
{"ㄓㄤ", IS_CHEWING|IS_PINYIN, 416},
{"ㄓㄥ", IS_CHEWING|IS_PINYIN, 421},
{"ㄓㄨ", IS_CHEWING|IS_PINYIN, 425},
{"ㄓㄨㄚ", IS_CHEWING|IS_PINYIN, 426},
{"ㄓㄨㄛ", IS_CHEWING|IS_PINYIN, 432},
{"ㄓㄨㄞ", IS_CHEWING|IS_PINYIN, 427},
{"ㄓㄨㄟ", IS_CHEWING|IS_PINYIN, 430},
{"ㄓㄨㄢ", IS_CHEWING|IS_PINYIN, 428},
{"ㄓㄨㄣ", IS_CHEWING|IS_PINYIN, 431},
{"ㄓㄨㄤ", IS_CHEWING|IS_PINYIN, 429},
{"ㄓㄨㄥ", IS_CHEWING|IS_PINYIN, 423},
{"ㄔ", IS_CHEWING|IS_PINYIN, 41},
{"ㄔㄚ", IS_CHEWING|IS_PINYIN, 33},
{"ㄔㄜ", IS_CHEWING|IS_PINYIN, 38},
{"ㄔㄞ", IS_CHEWING|IS_PINYIN, 34},
{"ㄔㄠ", IS_CHEWING|IS_PINYIN, 37},
{"ㄔㄡ", IS_CHEWING|IS_PINYIN, 43},
{"ㄔㄢ", IS_CHEWING|IS_PINYIN, 35},
{"ㄔㄣ", IS_CHEWING|IS_PINYIN, 39},
{"ㄔㄤ", IS_CHEWING|IS_PINYIN, 36},
{"ㄔㄥ", IS_CHEWING|IS_PINYIN, 40},
{"ㄔㄨ", IS_CHEWING|IS_PINYIN, 44},
{"ㄔㄨㄚ", IS_CHEWING, 45},
{"ㄔㄨㄛ", IS_CHEWING|IS_PINYIN, 51},
{"ㄔㄨㄞ", IS_CHEWING|IS_PINYIN, 46},
{"ㄔㄨㄟ", IS_CHEWING|IS_PINYIN, 49},
{"ㄔㄨㄢ", IS_CHEWING|IS_PINYIN, 47},
{"ㄔㄨㄣ", IS_CHEWING|IS_PINYIN, 50},
{"ㄔㄨㄤ", IS_CHEWING|IS_PINYIN, 48},
{"ㄔㄨㄥ", IS_CHEWING|IS_PINYIN, 42},
{"ㄕ", IS_CHEWING|IS_PINYIN, 322},
{"ㄕㄚ", IS_CHEWING|IS_PINYIN, 313},
{"ㄕㄜ", IS_CHEWING|IS_PINYIN, 318},
{"ㄕㄞ", IS_CHEWING|IS_PINYIN, 314},
{"ㄕㄟ", IS_CHEWING|IS_PINYIN, 319},
{"ㄕㄠ", IS_CHEWING|IS_PINYIN, 317},
{"ㄕㄡ", IS_CHEWING|IS_PINYIN, 323},
{"ㄕㄢ", IS_CHEWING|IS_PINYIN, 315},
{"ㄕㄣ", IS_CHEWING|IS_PINYIN, 320},
{"ㄕㄤ", IS_CHEWING|IS_PINYIN, 316},
{"ㄕㄥ", IS_CHEWING|IS_PINYIN, 321},
{"ㄕㄨ", IS_CHEWING|IS_PINYIN, 324},
{"ㄕㄨㄚ", IS_CHEWING|IS_PINYIN, 325},
{"ㄕㄨㄛ", IS_CHEWING|IS_PINYIN, 331},
{"ㄕㄨㄞ", IS_CHEWING|IS_PINYIN, 326},
{"ㄕㄨㄟ", IS_CHEWING|IS_PINYIN, 329},
{"ㄕㄨㄢ", IS_CHEWING|IS_PINYIN, 327},
{"ㄕㄨㄣ", IS_CHEWING|IS_PINYIN, 330},
{"ㄕㄨㄤ", IS_CHEWING|IS_PINYIN, 328},
{"ㄖ", IS_CHEWING|IS_PINYIN, 294},
{"ㄖㄜ", IS_CHEWING|IS_PINYIN, 291},
{"ㄖㄠ", IS_CHEWING|IS_PINYIN, 290},
{"ㄖㄡ", IS_CHEWING|IS_PINYIN, 296},
{"ㄖㄢ", IS_CHEWING|IS_PINYIN, 288},
{"ㄖㄣ", IS_CHEWING|IS_PINYIN, 292},
{"ㄖㄤ", IS_CHEWING|IS_PINYIN, 289},
{"ㄖㄥ", IS_CHEWING|IS_PINYIN, 293},
{"ㄖㄨ", IS_CHEWING|IS_PINYIN, 297},
{"ㄖㄨㄚ", IS_CHEWING, 298},
{"ㄖㄨㄛ", IS_CHEWING|IS_PINYIN, 302},
{"ㄖㄨㄟ", IS_CHEWING|IS_PINYIN, 300},
{"ㄖㄨㄢ", IS_CHEWING|IS_PINYIN, 299},
{"ㄖㄨㄣ", IS_CHEWING|IS_PINYIN, 301},
{"ㄖㄨㄥ", IS_CHEWING|IS_PINYIN, 295},
{"ㄗ", IS_CHEWING|IS_PINYIN, 433},
{"ㄗㄚ", IS_CHEWING|IS_PINYIN, 403},
{"ㄗㄜ", IS_CHEWING|IS_PINYIN, 408},
{"ㄗㄞ", IS_CHEWING|IS_PINYIN, 404},
{"ㄗㄟ", IS_CHEWING|IS_PINYIN, 409},
{"ㄗㄠ", IS_CHEWING|IS_PINYIN, 407},
{"ㄗㄡ", IS_CHEWING|IS_PINYIN, 435},
{"ㄗㄢ", IS_CHEWING|IS_PINYIN, 405},
{"ㄗㄣ", IS_CHEWING|IS_PINYIN, 410},
{"ㄗㄤ", IS_CHEWING|IS_PINYIN, 406},
{"ㄗㄥ", IS_CHEWING|IS_PINYIN, 411},
{"ㄗㄨ", IS_CHEWING|IS_PINYIN, 436},
{"ㄗㄨㄛ", IS_CHEWING|IS_PINYIN, 440},
{"ㄗㄨㄟ", IS_CHEWING|IS_PINYIN, 438},
{"ㄗㄨㄢ", IS_CHEWING|IS_PINYIN, 437},
{"ㄗㄨㄣ", IS_CHEWING|IS_PINYIN, 439},
{"ㄗㄨㄥ", IS_CHEWING|IS_PINYIN, 434},
{"ㄘ", IS_CHEWING|IS_PINYIN, 52},
{"ㄘㄚ", IS_CHEWING|IS_PINYIN, 24},
{"ㄘㄜ", IS_CHEWING|IS_PINYIN, 29},
{"ㄘㄞ", IS_CHEWING|IS_PINYIN, 25},
{"ㄘㄠ", IS_CHEWING|IS_PINYIN, 28},
{"ㄘㄡ", IS_CHEWING|IS_PINYIN, 54},
{"ㄘㄢ", IS_CHEWING|IS_PINYIN, 26},
{"ㄘㄣ", IS_CHEWING|IS_PINYIN, 30},
{"ㄘㄤ", IS_CHEWING|IS_PINYIN, 27},
{"ㄘㄥ", IS_CHEWING|IS_PINYIN, 31},
{"ㄘㄨ", IS_CHEWING|IS_PINYIN, 55},
{"ㄘㄨㄛ", IS_CHEWING|IS_PINYIN, 59},
{"ㄘㄨㄟ", IS_CHEWING|IS_PINYIN, 57},
{"ㄘㄨㄢ", IS_CHEWING|IS_PINYIN, 56},
{"ㄘㄨㄣ", IS_CHEWING|IS_PINYIN, 58},
{"ㄘㄨㄥ", IS_CHEWING|IS_PINYIN, 53},
{"ㄙ", IS_CHEWING|IS_PINYIN, 332},
{"ㄙㄚ", IS_CHEWING|IS_PINYIN, 304},
{"ㄙㄜ", IS_CHEWING|IS_PINYIN, 309},
{"ㄙㄞ", IS_CHEWING|IS_PINYIN, 305},
{"ㄙㄠ", IS_CHEWING|IS_PINYIN, 308},
{"ㄙㄡ", IS_CHEWING|IS_PINYIN, 334},
{"ㄙㄢ", IS_CHEWING|IS_PINYIN, 306},
{"ㄙㄣ", IS_CHEWING|IS_PINYIN, 310},
{"ㄙㄤ", IS_CHEWING|IS_PINYIN, 307},
{"ㄙㄥ", IS_CHEWING|IS_PINYIN, 311},
{"ㄙㄨ", IS_CHEWING|IS_PINYIN, 335},
{"ㄙㄨㄛ", IS_CHEWING|IS_PINYIN, 339},
{"ㄙㄨㄟ", IS_CHEWING|IS_PINYIN, 337},
{"ㄙㄨㄢ", IS_CHEWING|IS_PINYIN, 336},
{"ㄙㄨㄣ", IS_CHEWING|IS_PINYIN, 338},
{"ㄙㄨㄥ", IS_CHEWING|IS_PINYIN, 333},
{"ㄚ", IS_CHEWING|IS_PINYIN, 1},
{"ㄛ", IS_CHEWING|IS_PINYIN, 252},
{"ㄜ", IS_CHEWING|IS_PINYIN, 85},
{"ㄞ", IS_CHEWING|IS_PINYIN, 2},
{"ㄟ", IS_CHEWING|IS_PINYIN, 86},
{"ㄠ", IS_CHEWING|IS_PINYIN, 5},
{"ㄡ", IS_CHEWING|IS_PINYIN, 253},
{"ㄢ", IS_CHEWING|IS_PINYIN, 3},
{"ㄣ", IS_CHEWING|IS_PINYIN, 87},
{"ㄤ", IS_CHEWING|IS_PINYIN, 4},
{"ㄥ", IS_CHEWING, 88},
{"ㄦ", IS_CHEWING|IS_PINYIN, 89},
{"ㄧ", IS_CHEWING|IS_PINYIN, 392},
{"ㄧㄚ", IS_CHEWING|IS_PINYIN, 386},
{"ㄧㄛ", IS_CHEWING|IS_PINYIN, 395},
{"ㄧㄝ", IS_CHEWING|IS_PINYIN, 391},
{"ㄧㄞ", IS_CHEWING, 387},
{"ㄧㄠ", IS_CHEWING|IS_PINYIN, 390},
{"ㄧㄡ", IS_CHEWING|IS_PINYIN, 397},
{"ㄧㄢ", IS_CHEWING|IS_PINYIN, 388},
{"ㄧㄣ", IS_CHEWING|IS_PINYIN, 393},
{"ㄧㄤ", IS_CHEWING|IS_PINYIN, 389},
{"ㄧㄥ", IS_CHEWING|IS_PINYIN, 394},
{"ㄨ", IS_CHEWING|IS_PINYIN, 369},
{"ㄨㄚ", IS_CHEWING|IS_PINYIN, 361},
{"ㄨㄛ", IS_CHEWING|IS_PINYIN, 368},
{"ㄨㄞ", IS_CHEWING|IS_PINYIN, 362},
{"ㄨㄟ", IS_CHEWING|IS_PINYIN, 365},
{"ㄨㄢ", IS_CHEWING|IS_PINYIN, 363},
{"ㄨㄣ", IS_CHEWING|IS_PINYIN, 366},
{"ㄨㄤ", IS_CHEWING|IS_PINYIN, 364},
{"ㄨㄥ", IS_CHEWING|IS_PINYIN, 367},
{"ㄩ", IS_CHEWING|IS_PINYIN, 398},
{"ㄩㄝ", IS_CHEWING|IS_PINYIN, 400},
{"ㄩㄢ", IS_CHEWING|IS_PINYIN, 399},
{"ㄩㄣ", IS_CHEWING|IS_PINYIN, 401},
{"ㄩㄥ", IS_CHEWING|IS_PINYIN, 396},
{"ㄫ", IS_CHEWING|IS_PINYIN, 234}
};

const content_table_item_t content_table[] = {
{"", "", "", "", ChewingKey()},
{"a", "", "a", "ㄚ", ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"ai", "", "ai", "ㄞ", ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"an", "", "an", "ㄢ", ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"ang", "", "ang", "ㄤ", ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"ao", "", "ao", "ㄠ", ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"b", "b", "", "ㄅ", ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"ba", "b", "a", "ㄅㄚ", ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"bai", "b", "ai", "ㄅㄞ", ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"ban", "b", "an", "ㄅㄢ", ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"bang", "b", "ang", "ㄅㄤ", ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"bao", "b", "ao", "ㄅㄠ", ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"bei", "b", "ei", "ㄅㄟ", ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_EI)},
{"ben", "b", "en", "ㄅㄣ", ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"beng", "b", "eng", "ㄅㄥ", ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"bi", "b", "i", "ㄅㄧ", ChewingKey(CHEWING_B, CHEWING_I, CHEWING_ZERO_FINAL)},
{"bian", "b", "ian", "ㄅㄧㄢ", ChewingKey(CHEWING_B, CHEWING_I, CHEWING_AN)},
{"biao", "b", "iao", "ㄅㄧㄠ", ChewingKey(CHEWING_B, CHEWING_I, CHEWING_AO)},
{"bie", "b", "ie", "ㄅㄧㄝ", ChewingKey(CHEWING_B, CHEWING_I, CHEWING_E)},
{"bin", "b", "in", "ㄅㄧㄣ", ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, PINYIN_IN)},
{"bing", "b", "ing", "ㄅㄧㄥ", ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, PINYIN_ING)},
{"bo", "b", "o", "ㄅㄛ", ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_O)},
{"bu", "b", "u", "ㄅㄨ", ChewingKey(CHEWING_B, CHEWING_U, CHEWING_ZERO_FINAL)},
{"c", "c", "", "ㄘ", ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"ca", "c", "a", "ㄘㄚ", ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"cai", "c", "ai", "ㄘㄞ", ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"can", "c", "an", "ㄘㄢ", ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"cang", "c", "ang", "ㄘㄤ", ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"cao", "c", "ao", "ㄘㄠ", ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"ce", "c", "e", "ㄘㄜ", ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"cen", "c", "en", "ㄘㄣ", ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"ceng", "c", "eng", "ㄘㄥ", ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"ch", "ch", "", "ㄔ", ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"cha", "ch", "a", "ㄔㄚ", ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"chai", "ch", "ai", "ㄔㄞ", ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"chan", "ch", "an", "ㄔㄢ", ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"chang", "ch", "ang", "ㄔㄤ", ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"chao", "ch", "ao", "ㄔㄠ", ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"che", "ch", "e", "ㄔㄜ", ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"chen", "ch", "en", "ㄔㄣ", ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"cheng", "ch", "eng", "ㄔㄥ", ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"chi", "ch", "i", "ㄔ", ChewingKey(CHEWING_CH, CHEWING_I, CHEWING_ZERO_FINAL)},
{"chong", "ch", "ong", "ㄔㄨㄥ", ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, PINYIN_ONG)},
{"chou", "ch", "ou", "ㄔㄡ", ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"chu", "ch", "u", "ㄔㄨ", ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_ZERO_FINAL)},
{"chua", "ch", "ua", "ㄔㄨㄚ", ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_A)},
{"chuai", "ch", "uai", "ㄔㄨㄞ", ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_AI)},
{"chuan", "ch", "uan", "ㄔㄨㄢ", ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_AN)},
{"chuang", "ch", "uang", "ㄔㄨㄤ", ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_ANG)},
{"chui", "ch", "ui", "ㄔㄨㄟ", ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_EI)},
{"chun", "ch", "un", "ㄔㄨㄣ", ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_EN)},
{"chuo", "ch", "uo", "ㄔㄨㄛ", ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_O)},
{"ci", "c", "i", "ㄘ", ChewingKey(CHEWING_C, CHEWING_I, CHEWING_ZERO_FINAL)},
{"cong", "c", "ong", "ㄘㄨㄥ", ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, PINYIN_ONG)},
{"cou", "c", "ou", "ㄘㄡ", ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"cu", "c", "u", "ㄘㄨ", ChewingKey(CHEWING_C, CHEWING_U, CHEWING_ZERO_FINAL)},
{"cuan", "c", "uan", "ㄘㄨㄢ", ChewingKey(CHEWING_C, CHEWING_U, CHEWING_AN)},
{"cui", "c", "ui", "ㄘㄨㄟ", ChewingKey(CHEWING_C, CHEWING_U, CHEWING_EI)},
{"cun", "c", "un", "ㄘㄨㄣ", ChewingKey(CHEWING_C, CHEWING_U, CHEWING_EN)},
{"cuo", "c", "uo", "ㄘㄨㄛ", ChewingKey(CHEWING_C, CHEWING_U, CHEWING_O)},
{"d", "d", "", "ㄉ", ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"da", "d", "a", "ㄉㄚ", ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"dai", "d", "ai", "ㄉㄞ", ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"dan", "d", "an", "ㄉㄢ", ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"dang", "d", "ang", "ㄉㄤ", ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"dao", "d", "ao", "ㄉㄠ", ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"de", "d", "e", "ㄉㄜ", ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"dei", "d", "ei", "ㄉㄟ", ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_EI)},
{"den", "d", "en", "ㄉㄣ", ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"deng", "d", "eng", "ㄉㄥ", ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"di", "d", "i", "ㄉㄧ", ChewingKey(CHEWING_D, CHEWING_I, CHEWING_ZERO_FINAL)},
{"dia", "d", "ia", "ㄉㄧㄚ", ChewingKey(CHEWING_D, CHEWING_I, CHEWING_A)},
{"dian", "d", "ian", "ㄉㄧㄢ", ChewingKey(CHEWING_D, CHEWING_I, CHEWING_AN)},
{"diao", "d", "iao", "ㄉㄧㄠ", ChewingKey(CHEWING_D, CHEWING_I, CHEWING_AO)},
{"die", "d", "ie", "ㄉㄧㄝ", ChewingKey(CHEWING_D, CHEWING_I, CHEWING_E)},
{"din", "d", "in", "ㄉㄧㄣ", ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, PINYIN_IN)},
{"ding", "d", "ing", "ㄉㄧㄥ", ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, PINYIN_ING)},
{"diu", "d", "iu", "ㄉㄧㄡ", ChewingKey(CHEWING_D, CHEWING_I, CHEWING_OU)},
{"dong", "d", "ong", "ㄉㄨㄥ", ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, PINYIN_ONG)},
{"dou", "d", "ou", "ㄉㄡ", ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"du", "d", "u", "ㄉㄨ", ChewingKey(CHEWING_D, CHEWING_U, CHEWING_ZERO_FINAL)},
{"duan", "d", "uan", "ㄉㄨㄢ", ChewingKey(CHEWING_D, CHEWING_U, CHEWING_AN)},
{"dui", "d", "ui", "ㄉㄨㄟ", ChewingKey(CHEWING_D, CHEWING_U, CHEWING_EI)},
{"dun", "d", "un", "ㄉㄨㄣ", ChewingKey(CHEWING_D, CHEWING_U, CHEWING_EN)},
{"duo", "d", "uo", "ㄉㄨㄛ", ChewingKey(CHEWING_D, CHEWING_U, CHEWING_O)},
{"e", "", "e", "ㄜ", ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"ei", "", "ei", "ㄟ", ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_EI)},
{"en", "", "en", "ㄣ", ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"eng", "", "eng", "ㄥ", ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"er", "", "er", "ㄦ", ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_ER)},
{"f", "f", "", "ㄈ", ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"fa", "f", "a", "ㄈㄚ", ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"fan", "f", "an", "ㄈㄢ", ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"fang", "f", "ang", "ㄈㄤ", ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"fe", "f", "e", "ㄈㄜ", ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"fei", "f", "ei", "ㄈㄟ", ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_EI)},
{"fen", "f", "en", "ㄈㄣ", ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"feng", "f", "eng", "ㄈㄥ", ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"fo", "f", "o", "ㄈㄛ", ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_O)},
{"fou", "f", "ou", "ㄈㄡ", ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"fu", "f", "u", "ㄈㄨ", ChewingKey(CHEWING_F, CHEWING_U, CHEWING_ZERO_FINAL)},
{"g", "g", "", "ㄍ", ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"ga", "g", "a", "ㄍㄚ", ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"gai", "g", "ai", "ㄍㄞ", ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"gan", "g", "an", "ㄍㄢ", ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"gang", "g", "ang", "ㄍㄤ", ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"gao", "g", "ao", "ㄍㄠ", ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"ge", "g", "e", "ㄍㄜ", ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"gei", "g", "ei", "ㄍㄟ", ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_EI)},
{"gen", "g", "en", "ㄍㄣ", ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"geng", "g", "eng", "ㄍㄥ", ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"gong", "g", "ong", "ㄍㄨㄥ", ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, PINYIN_ONG)},
{"gou", "g", "ou", "ㄍㄡ", ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"gu", "g", "u", "ㄍㄨ", ChewingKey(CHEWING_G, CHEWING_U, CHEWING_ZERO_FINAL)},
{"gua", "g", "ua", "ㄍㄨㄚ", ChewingKey(CHEWING_G, CHEWING_U, CHEWING_A)},
{"guai", "g", "uai", "ㄍㄨㄞ", ChewingKey(CHEWING_G, CHEWING_U, CHEWING_AI)},
{"guan", "g", "uan", "ㄍㄨㄢ", ChewingKey(CHEWING_G, CHEWING_U, CHEWING_AN)},
{"guang", "g", "uang", "ㄍㄨㄤ", ChewingKey(CHEWING_G, CHEWING_U, CHEWING_ANG)},
{"gui", "g", "ui", "ㄍㄨㄟ", ChewingKey(CHEWING_G, CHEWING_U, CHEWING_EI)},
{"gun", "g", "un", "ㄍㄨㄣ", ChewingKey(CHEWING_G, CHEWING_U, CHEWING_EN)},
{"guo", "g", "uo", "ㄍㄨㄛ", ChewingKey(CHEWING_G, CHEWING_U, CHEWING_O)},
{"h", "h", "", "ㄏ", ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"ha", "h", "a", "ㄏㄚ", ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"hai", "h", "ai", "ㄏㄞ", ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"han", "h", "an", "ㄏㄢ", ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"hang", "h", "ang", "ㄏㄤ", ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"hao", "h", "ao", "ㄏㄠ", ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"he", "h", "e", "ㄏㄜ", ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"hei", "h", "ei", "ㄏㄟ", ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_EI)},
{"hen", "h", "en", "ㄏㄣ", ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"heng", "h", "eng", "ㄏㄥ", ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"hong", "h", "ong", "ㄏㄨㄥ", ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, PINYIN_ONG)},
{"hou", "h", "ou", "ㄏㄡ", ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"hu", "h", "u", "ㄏㄨ", ChewingKey(CHEWING_H, CHEWING_U, CHEWING_ZERO_FINAL)},
{"hua", "h", "ua", "ㄏㄨㄚ", ChewingKey(CHEWING_H, CHEWING_U, CHEWING_A)},
{"huai", "h", "uai", "ㄏㄨㄞ", ChewingKey(CHEWING_H, CHEWING_U, CHEWING_AI)},
{"huan", "h", "uan", "ㄏㄨㄢ", ChewingKey(CHEWING_H, CHEWING_U, CHEWING_AN)},
{"huang", "h", "uang", "ㄏㄨㄤ", ChewingKey(CHEWING_H, CHEWING_U, CHEWING_ANG)},
{"hui", "h", "ui", "ㄏㄨㄟ", ChewingKey(CHEWING_H, CHEWING_U, CHEWING_EI)},
{"hun", "h", "un", "ㄏㄨㄣ", ChewingKey(CHEWING_H, CHEWING_U, CHEWING_EN)},
{"huo", "h", "uo", "ㄏㄨㄛ", ChewingKey(CHEWING_H, CHEWING_U, CHEWING_O)},
{"j", "j", "", "ㄐ", ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"ji", "j", "i", "ㄐㄧ", ChewingKey(CHEWING_J, CHEWING_I, CHEWING_ZERO_FINAL)},
{"jia", "j", "ia", "ㄐㄧㄚ", ChewingKey(CHEWING_J, CHEWING_I, CHEWING_A)},
{"jian", "j", "ian", "ㄐㄧㄢ", ChewingKey(CHEWING_J, CHEWING_I, CHEWING_AN)},
{"jiang", "j", "iang", "ㄐㄧㄤ", ChewingKey(CHEWING_J, CHEWING_I, CHEWING_ANG)},
{"jiao", "j", "iao", "ㄐㄧㄠ", ChewingKey(CHEWING_J, CHEWING_I, CHEWING_AO)},
{"jie", "j", "ie", "ㄐㄧㄝ", ChewingKey(CHEWING_J, CHEWING_I, CHEWING_E)},
{"jin", "j", "in", "ㄐㄧㄣ", ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, PINYIN_IN)},
{"jing", "j", "ing", "ㄐㄧㄥ", ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, PINYIN_ING)},
{"jiong", "j", "iong", "ㄐㄩㄥ", ChewingKey(CHEWING_J, CHEWING_I, PINYIN_ONG)},
{"jiu", "j", "iu", "ㄐㄧㄡ", ChewingKey(CHEWING_J, CHEWING_I, CHEWING_OU)},
{"ju", "j", "u", "ㄐㄩ", ChewingKey(CHEWING_J, CHEWING_V, CHEWING_ZERO_FINAL)},
{"juan", "j", "uan", "ㄐㄩㄢ", ChewingKey(CHEWING_J, CHEWING_V, CHEWING_AN)},
{"jue", "j", "ue", "ㄐㄩㄝ", ChewingKey(CHEWING_J, CHEWING_V, CHEWING_E)},
{"jun", "j", "un", "ㄐㄩㄣ", ChewingKey(CHEWING_J, CHEWING_V, CHEWING_EN)},
{"k", "k", "", "ㄎ", ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"ka", "k", "a", "ㄎㄚ", ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"kai", "k", "ai", "ㄎㄞ", ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"kan", "k", "an", "ㄎㄢ", ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"kang", "k", "ang", "ㄎㄤ", ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"kao", "k", "ao", "ㄎㄠ", ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"ke", "k", "e", "ㄎㄜ", ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"kei", "k", "ei", "ㄎㄟ", ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_EI)},
{"ken", "k", "en", "ㄎㄣ", ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"keng", "k", "eng", "ㄎㄥ", ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"kong", "k", "ong", "ㄎㄨㄥ", ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, PINYIN_ONG)},
{"kou", "k", "ou", "ㄎㄡ", ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"ku", "k", "u", "ㄎㄨ", ChewingKey(CHEWING_K, CHEWING_U, CHEWING_ZERO_FINAL)},
{"kua", "k", "ua", "ㄎㄨㄚ", ChewingKey(CHEWING_K, CHEWING_U, CHEWING_A)},
{"kuai", "k", "uai", "ㄎㄨㄞ", ChewingKey(CHEWING_K, CHEWING_U, CHEWING_AI)},
{"kuan", "k", "uan", "ㄎㄨㄢ", ChewingKey(CHEWING_K, CHEWING_U, CHEWING_AN)},
{"kuang", "k", "uang", "ㄎㄨㄤ", ChewingKey(CHEWING_K, CHEWING_U, CHEWING_ANG)},
{"kui", "k", "ui", "ㄎㄨㄟ", ChewingKey(CHEWING_K, CHEWING_U, CHEWING_EI)},
{"kun", "k", "un", "ㄎㄨㄣ", ChewingKey(CHEWING_K, CHEWING_U, CHEWING_EN)},
{"kuo", "k", "uo", "ㄎㄨㄛ", ChewingKey(CHEWING_K, CHEWING_U, CHEWING_O)},
{"l", "l", "", "ㄌ", ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"la", "l", "a", "ㄌㄚ", ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"lai", "l", "ai", "ㄌㄞ", ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"lan", "l", "an", "ㄌㄢ", ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"lang", "l", "ang", "ㄌㄤ", ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"lao", "l", "ao", "ㄌㄠ", ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"le", "l", "e", "ㄌㄜ", ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"lei", "l", "ei", "ㄌㄟ", ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_EI)},
{"len", "l", "en", "ㄌㄣ", ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"leng", "l", "eng", "ㄌㄥ", ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"li", "l", "i", "ㄌㄧ", ChewingKey(CHEWING_L, CHEWING_I, CHEWING_ZERO_FINAL)},
{"lia", "l", "ia", "ㄌㄧㄚ", ChewingKey(CHEWING_L, CHEWING_I, CHEWING_A)},
{"lian", "l", "ian", "ㄌㄧㄢ", ChewingKey(CHEWING_L, CHEWING_I, CHEWING_AN)},
{"liang", "l", "iang", "ㄌㄧㄤ", ChewingKey(CHEWING_L, CHEWING_I, CHEWING_ANG)},
{"liao", "l", "iao", "ㄌㄧㄠ", ChewingKey(CHEWING_L, CHEWING_I, CHEWING_AO)},
{"lie", "l", "ie", "ㄌㄧㄝ", ChewingKey(CHEWING_L, CHEWING_I, CHEWING_E)},
{"lin", "l", "in", "ㄌㄧㄣ", ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, PINYIN_IN)},
{"ling", "l", "ing", "ㄌㄧㄥ", ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, PINYIN_ING)},
{"liu", "l", "iu", "ㄌㄧㄡ", ChewingKey(CHEWING_L, CHEWING_I, CHEWING_OU)},
{"lo", "l", "o", "ㄌㄛ", ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_O)},
{"long", "l", "ong", "ㄌㄨㄥ", ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, PINYIN_ONG)},
{"lou", "l", "ou", "ㄌㄡ", ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"lu", "l", "u", "ㄌㄨ", ChewingKey(CHEWING_L, CHEWING_U, CHEWING_ZERO_FINAL)},
{"luan", "l", "uan", "ㄌㄨㄢ", ChewingKey(CHEWING_L, CHEWING_U, CHEWING_AN)},
{"lun", "l", "un", "ㄌㄨㄣ", ChewingKey(CHEWING_L, CHEWING_U, CHEWING_EN)},
{"luo", "l", "uo", "ㄌㄨㄛ", ChewingKey(CHEWING_L, CHEWING_U, CHEWING_O)},
{"lv", "l", "v", "ㄌㄩ", ChewingKey(CHEWING_L, CHEWING_V, CHEWING_ZERO_FINAL)},
{"lve", "l", "ve", "ㄌㄩㄝ", ChewingKey(CHEWING_L, CHEWING_V, CHEWING_E)},
{"m", "m", "", "ㄇ", ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"ma", "m", "a", "ㄇㄚ", ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"mai", "m", "ai", "ㄇㄞ", ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"man", "m", "an", "ㄇㄢ", ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"mang", "m", "ang", "ㄇㄤ", ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"mao", "m", "ao", "ㄇㄠ", ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"me", "m", "e", "ㄇㄜ", ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"mei", "m", "ei", "ㄇㄟ", ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_EI)},
{"men", "m", "en", "ㄇㄣ", ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"meng", "m", "eng", "ㄇㄥ", ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"mi", "m", "i", "ㄇㄧ", ChewingKey(CHEWING_M, CHEWING_I, CHEWING_ZERO_FINAL)},
{"mian", "m", "ian", "ㄇㄧㄢ", ChewingKey(CHEWING_M, CHEWING_I, CHEWING_AN)},
{"miao", "m", "iao", "ㄇㄧㄠ", ChewingKey(CHEWING_M, CHEWING_I, CHEWING_AO)},
{"mie", "m", "ie", "ㄇㄧㄝ", ChewingKey(CHEWING_M, CHEWING_I, CHEWING_E)},
{"min", "m", "in", "ㄇㄧㄣ", ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, PINYIN_IN)},
{"ming", "m", "ing", "ㄇㄧㄥ", ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, PINYIN_ING)},
{"miu", "m", "iu", "ㄇㄧㄡ", ChewingKey(CHEWING_M, CHEWING_I, CHEWING_OU)},
{"mo", "m", "o", "ㄇㄛ", ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_O)},
{"mou", "m", "ou", "ㄇㄡ", ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"mu", "m", "u", "ㄇㄨ", ChewingKey(CHEWING_M, CHEWING_U, CHEWING_ZERO_FINAL)},
{"n", "n", "", "ㄋ", ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"na", "n", "a", "ㄋㄚ", ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"nai", "n", "ai", "ㄋㄞ", ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"nan", "n", "an", "ㄋㄢ", ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"nang", "n", "ang", "ㄋㄤ", ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"nao", "n", "ao", "ㄋㄠ", ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"ne", "n", "e", "ㄋㄜ", ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"nei", "n", "ei", "ㄋㄟ", ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_EI)},
{"nen", "n", "en", "ㄋㄣ", ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"neng", "n", "eng", "ㄋㄥ", ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"ng", "", "ng", "ㄫ", ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_NG)},
{"ni", "n", "i", "ㄋㄧ", ChewingKey(CHEWING_N, CHEWING_I, CHEWING_ZERO_FINAL)},
{"nia", "n", "ia", "ㄋㄧㄚ", ChewingKey(CHEWING_N, CHEWING_I, CHEWING_A)},
{"nian", "n", "ian", "ㄋㄧㄢ", ChewingKey(CHEWING_N, CHEWING_I, CHEWING_AN)},
{"niang", "n", "iang", "ㄋㄧㄤ", ChewingKey(CHEWING_N, CHEWING_I, CHEWING_ANG)},
{"niao", "n", "iao", "ㄋㄧㄠ", ChewingKey(CHEWING_N, CHEWING_I, CHEWING_AO)},
{"nie", "n", "ie", "ㄋㄧㄝ", ChewingKey(CHEWING_N, CHEWING_I, CHEWING_E)},
{"nin", "n", "in", "ㄋㄧㄣ", ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, PINYIN_IN)},
{"ning", "n", "ing", "ㄋㄧㄥ", ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, PINYIN_ING)},
{"niu", "n", "iu", "ㄋㄧㄡ", ChewingKey(CHEWING_N, CHEWING_I, CHEWING_OU)},
{"nong", "n", "ong", "ㄋㄨㄥ", ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, PINYIN_ONG)},
{"nou", "n", "ou", "ㄋㄡ", ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"nu", "n", "u", "ㄋㄨ", ChewingKey(CHEWING_N, CHEWING_U, CHEWING_ZERO_FINAL)},
{"nuan", "n", "uan", "ㄋㄨㄢ", ChewingKey(CHEWING_N, CHEWING_U, CHEWING_AN)},
{"nun", "n", "un", "ㄋㄨㄣ", ChewingKey(CHEWING_N, CHEWING_U, CHEWING_EN)},
{"nuo", "n", "uo", "ㄋㄨㄛ", ChewingKey(CHEWING_N, CHEWING_U, CHEWING_O)},
{"nv", "n", "v", "ㄋㄩ", ChewingKey(CHEWING_N, CHEWING_V, CHEWING_ZERO_FINAL)},
{"nve", "n", "ve", "ㄋㄩㄝ", ChewingKey(CHEWING_N, CHEWING_V, CHEWING_E)},
{"o", "", "o", "ㄛ", ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_O)},
{"ou", "", "ou", "ㄡ", ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"p", "p", "", "ㄆ", ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"pa", "p", "a", "ㄆㄚ", ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"pai", "p", "ai", "ㄆㄞ", ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"pan", "p", "an", "ㄆㄢ", ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"pang", "p", "ang", "ㄆㄤ", ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"pao", "p", "ao", "ㄆㄠ", ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"pei", "p", "ei", "ㄆㄟ", ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_EI)},
{"pen", "p", "en", "ㄆㄣ", ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"peng", "p", "eng", "ㄆㄥ", ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"pi", "p", "i", "ㄆㄧ", ChewingKey(CHEWING_P, CHEWING_I, CHEWING_ZERO_FINAL)},
{"pian", "p", "ian", "ㄆㄧㄢ", ChewingKey(CHEWING_P, CHEWING_I, CHEWING_AN)},
{"piao", "p", "iao", "ㄆㄧㄠ", ChewingKey(CHEWING_P, CHEWING_I, CHEWING_AO)},
{"pie", "p", "ie", "ㄆㄧㄝ", ChewingKey(CHEWING_P, CHEWING_I, CHEWING_E)},
{"pin", "p", "in", "ㄆㄧㄣ", ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, PINYIN_IN)},
{"ping", "p", "ing", "ㄆㄧㄥ", ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, PINYIN_ING)},
{"po", "p", "o", "ㄆㄛ", ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_O)},
{"pou", "p", "ou", "ㄆㄡ", ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"pu", "p", "u", "ㄆㄨ", ChewingKey(CHEWING_P, CHEWING_U, CHEWING_ZERO_FINAL)},
{"q", "q", "", "ㄑ", ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"qi", "q", "i", "ㄑㄧ", ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_ZERO_FINAL)},
{"qia", "q", "ia", "ㄑㄧㄚ", ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_A)},
{"qian", "q", "ian", "ㄑㄧㄢ", ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_AN)},
{"qiang", "q", "iang", "ㄑㄧㄤ", ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_ANG)},
{"qiao", "q", "iao", "ㄑㄧㄠ", ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_AO)},
{"qie", "q", "ie", "ㄑㄧㄝ", ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_E)},
{"qin", "q", "in", "ㄑㄧㄣ", ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, PINYIN_IN)},
{"qing", "q", "ing", "ㄑㄧㄥ", ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, PINYIN_ING)},
{"qiong", "q", "iong", "ㄑㄩㄥ", ChewingKey(CHEWING_Q, CHEWING_I, PINYIN_ONG)},
{"qiu", "q", "iu", "ㄑㄧㄡ", ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_OU)},
{"qu", "q", "u", "ㄑㄩ", ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_ZERO_FINAL)},
{"quan", "q", "uan", "ㄑㄩㄢ", ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_AN)},
{"que", "q", "ue", "ㄑㄩㄝ", ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_E)},
{"qun", "q", "un", "ㄑㄩㄣ", ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_EN)},
{"r", "r", "", "ㄖ", ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"ran", "r", "an", "ㄖㄢ", ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"rang", "r", "ang", "ㄖㄤ", ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"rao", "r", "ao", "ㄖㄠ", ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"re", "r", "e", "ㄖㄜ", ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"ren", "r", "en", "ㄖㄣ", ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"reng", "r", "eng", "ㄖㄥ", ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"ri", "r", "i", "ㄖ", ChewingKey(CHEWING_R, CHEWING_I, CHEWING_ZERO_FINAL)},
{"rong", "r", "ong", "ㄖㄨㄥ", ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, PINYIN_ONG)},
{"rou", "r", "ou", "ㄖㄡ", ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"ru", "r", "u", "ㄖㄨ", ChewingKey(CHEWING_R, CHEWING_U, CHEWING_ZERO_FINAL)},
{"rua", "r", "ua", "ㄖㄨㄚ", ChewingKey(CHEWING_R, CHEWING_U, CHEWING_A)},
{"ruan", "r", "uan", "ㄖㄨㄢ", ChewingKey(CHEWING_R, CHEWING_U, CHEWING_AN)},
{"rui", "r", "ui", "ㄖㄨㄟ", ChewingKey(CHEWING_R, CHEWING_U, CHEWING_EI)},
{"run", "r", "un", "ㄖㄨㄣ", ChewingKey(CHEWING_R, CHEWING_U, CHEWING_EN)},
{"ruo", "r", "uo", "ㄖㄨㄛ", ChewingKey(CHEWING_R, CHEWING_U, CHEWING_O)},
{"s", "s", "", "ㄙ", ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"sa", "s", "a", "ㄙㄚ", ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"sai", "s", "ai", "ㄙㄞ", ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"san", "s", "an", "ㄙㄢ", ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"sang", "s", "ang", "ㄙㄤ", ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"sao", "s", "ao", "ㄙㄠ", ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"se", "s", "e", "ㄙㄜ", ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"sen", "s", "en", "ㄙㄣ", ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"seng", "s", "eng", "ㄙㄥ", ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"sh", "sh", "", "ㄕ", ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"sha", "sh", "a", "ㄕㄚ", ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"shai", "sh", "ai", "ㄕㄞ", ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"shan", "sh", "an", "ㄕㄢ", ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"shang", "sh", "ang", "ㄕㄤ", ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"shao", "sh", "ao", "ㄕㄠ", ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"she", "sh", "e", "ㄕㄜ", ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"shei", "sh", "ei", "ㄕㄟ", ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_EI)},
{"shen", "sh", "en", "ㄕㄣ", ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"sheng", "sh", "eng", "ㄕㄥ", ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"shi", "sh", "i", "ㄕ", ChewingKey(CHEWING_SH, CHEWING_I, CHEWING_ZERO_FINAL)},
{"shou", "sh", "ou", "ㄕㄡ", ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"shu", "sh", "u", "ㄕㄨ", ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_ZERO_FINAL)},
{"shua", "sh", "ua", "ㄕㄨㄚ", ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_A)},
{"shuai", "sh", "uai", "ㄕㄨㄞ", ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_AI)},
{"shuan", "sh", "uan", "ㄕㄨㄢ", ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_AN)},
{"shuang", "sh", "uang", "ㄕㄨㄤ", ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_ANG)},
{"shui", "sh", "ui", "ㄕㄨㄟ", ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_EI)},
{"shun", "sh", "un", "ㄕㄨㄣ", ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_EN)},
{"shuo", "sh", "uo", "ㄕㄨㄛ", ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_O)},
{"si", "s", "i", "ㄙ", ChewingKey(CHEWING_S, CHEWING_I, CHEWING_ZERO_FINAL)},
{"song", "s", "ong", "ㄙㄨㄥ", ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, PINYIN_ONG)},
{"sou", "s", "ou", "ㄙㄡ", ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"su", "s", "u", "ㄙㄨ", ChewingKey(CHEWING_S, CHEWING_U, CHEWING_ZERO_FINAL)},
{"suan", "s", "uan", "ㄙㄨㄢ", ChewingKey(CHEWING_S, CHEWING_U, CHEWING_AN)},
{"sui", "s", "ui", "ㄙㄨㄟ", ChewingKey(CHEWING_S, CHEWING_U, CHEWING_EI)},
{"sun", "s", "un", "ㄙㄨㄣ", ChewingKey(CHEWING_S, CHEWING_U, CHEWING_EN)},
{"suo", "s", "uo", "ㄙㄨㄛ", ChewingKey(CHEWING_S, CHEWING_U, CHEWING_O)},
{"t", "t", "", "ㄊ", ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"ta", "t", "a", "ㄊㄚ", ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"tai", "t", "ai", "ㄊㄞ", ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"tan", "t", "an", "ㄊㄢ", ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"tang", "t", "ang", "ㄊㄤ", ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"tao", "t", "ao", "ㄊㄠ", ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"te", "t", "e", "ㄊㄜ", ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"teng", "t", "eng", "ㄊㄥ", ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"ti", "t", "i", "ㄊㄧ", ChewingKey(CHEWING_T, CHEWING_I, CHEWING_ZERO_FINAL)},
{"tian", "t", "ian", "ㄊㄧㄢ", ChewingKey(CHEWING_T, CHEWING_I, CHEWING_AN)},
{"tiao", "t", "iao", "ㄊㄧㄠ", ChewingKey(CHEWING_T, CHEWING_I, CHEWING_AO)},
{"tie", "t", "ie", "ㄊㄧㄝ", ChewingKey(CHEWING_T, CHEWING_I, CHEWING_E)},
{"ting", "t", "ing", "ㄊㄧㄥ", ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, PINYIN_ING)},
{"tong", "t", "ong", "ㄊㄨㄥ", ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, PINYIN_ONG)},
{"tou", "t", "ou", "ㄊㄡ", ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"tu", "t", "u", "ㄊㄨ", ChewingKey(CHEWING_T, CHEWING_U, CHEWING_ZERO_FINAL)},
{"tuan", "t", "uan", "ㄊㄨㄢ", ChewingKey(CHEWING_T, CHEWING_U, CHEWING_AN)},
{"tui", "t", "ui", "ㄊㄨㄟ", ChewingKey(CHEWING_T, CHEWING_U, CHEWING_EI)},
{"tun", "t", "un", "ㄊㄨㄣ", ChewingKey(CHEWING_T, CHEWING_U, CHEWING_EN)},
{"tuo", "t", "uo", "ㄊㄨㄛ", ChewingKey(CHEWING_T, CHEWING_U, CHEWING_O)},
{"w", "w", "", "PINYIN_W", ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"wa", "w", "a", "ㄨㄚ", ChewingKey(PINYIN_W, CHEWING_U, CHEWING_A)},
{"wai", "w", "ai", "ㄨㄞ", ChewingKey(PINYIN_W, CHEWING_U, CHEWING_AI)},
{"wan", "w", "an", "ㄨㄢ", ChewingKey(PINYIN_W, CHEWING_U, CHEWING_AN)},
{"wang", "w", "ang", "ㄨㄤ", ChewingKey(PINYIN_W, CHEWING_U, CHEWING_ANG)},
{"wei", "w", "ei", "ㄨㄟ", ChewingKey(PINYIN_W, CHEWING_U, CHEWING_EI)},
{"wen", "w", "en", "ㄨㄣ", ChewingKey(PINYIN_W, CHEWING_U, CHEWING_EN)},
{"weng", "w", "eng", "ㄨㄥ", ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, PINYIN_ONG)},
{"wo", "w", "o", "ㄨㄛ", ChewingKey(PINYIN_W, CHEWING_U, CHEWING_O)},
{"wu", "w", "u", "ㄨ", ChewingKey(PINYIN_W, CHEWING_U, CHEWING_ZERO_FINAL)},
{"x", "x", "", "ㄒ", ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"xi", "x", "i", "ㄒㄧ", ChewingKey(CHEWING_X, CHEWING_I, CHEWING_ZERO_FINAL)},
{"xia", "x", "ia", "ㄒㄧㄚ", ChewingKey(CHEWING_X, CHEWING_I, CHEWING_A)},
{"xian", "x", "ian", "ㄒㄧㄢ", ChewingKey(CHEWING_X, CHEWING_I, CHEWING_AN)},
{"xiang", "x", "iang", "ㄒㄧㄤ", ChewingKey(CHEWING_X, CHEWING_I, CHEWING_ANG)},
{"xiao", "x", "iao", "ㄒㄧㄠ", ChewingKey(CHEWING_X, CHEWING_I, CHEWING_AO)},
{"xie", "x", "ie", "ㄒㄧㄝ", ChewingKey(CHEWING_X, CHEWING_I, CHEWING_E)},
{"xin", "x", "in", "ㄒㄧㄣ", ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, PINYIN_IN)},
{"xing", "x", "ing", "ㄒㄧㄥ", ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, PINYIN_ING)},
{"xiong", "x", "iong", "ㄒㄩㄥ", ChewingKey(CHEWING_X, CHEWING_I, PINYIN_ONG)},
{"xiu", "x", "iu", "ㄒㄧㄡ", ChewingKey(CHEWING_X, CHEWING_I, CHEWING_OU)},
{"xu", "x", "u", "ㄒㄩ", ChewingKey(CHEWING_X, CHEWING_V, CHEWING_ZERO_FINAL)},
{"xuan", "x", "uan", "ㄒㄩㄢ", ChewingKey(CHEWING_X, CHEWING_V, CHEWING_AN)},
{"xue", "x", "ue", "ㄒㄩㄝ", ChewingKey(CHEWING_X, CHEWING_V, CHEWING_E)},
{"xun", "x", "un", "ㄒㄩㄣ", ChewingKey(CHEWING_X, CHEWING_V, CHEWING_EN)},
{"y", "y", "", "PINYIN_Y", ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"ya", "y", "a", "ㄧㄚ", ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_A)},
{"yai", "y", "ai", "ㄧㄞ", ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_AI)},
{"yan", "y", "an", "ㄧㄢ", ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_AN)},
{"yang", "y", "ang", "ㄧㄤ", ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_ANG)},
{"yao", "y", "ao", "ㄧㄠ", ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_AO)},
{"ye", "y", "e", "ㄧㄝ", ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_E)},
{"yi", "y", "i", "ㄧ", ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_ZERO_FINAL)},
{"yin", "y", "in", "ㄧㄣ", ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, PINYIN_IN)},
{"ying", "y", "ing", "ㄧㄥ", ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, PINYIN_ING)},
{"yo", "y", "o", "ㄧㄛ", ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_O)},
{"yong", "y", "ong", "ㄩㄥ", ChewingKey(PINYIN_Y, CHEWING_I, PINYIN_ONG)},
{"you", "y", "ou", "ㄧㄡ", ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_OU)},
{"yu", "y", "u", "ㄩ", ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_ZERO_FINAL)},
{"yuan", "y", "uan", "ㄩㄢ", ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_AN)},
{"yue", "y", "ue", "ㄩㄝ", ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_E)},
{"yun", "y", "un", "ㄩㄣ", ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_EN)},
{"z", "z", "", "ㄗ", ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"za", "z", "a", "ㄗㄚ", ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"zai", "z", "ai", "ㄗㄞ", ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"zan", "z", "an", "ㄗㄢ", ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"zang", "z", "ang", "ㄗㄤ", ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"zao", "z", "ao", "ㄗㄠ", ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"ze", "z", "e", "ㄗㄜ", ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"zei", "z", "ei", "ㄗㄟ", ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_EI)},
{"zen", "z", "en", "ㄗㄣ", ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"zeng", "z", "eng", "ㄗㄥ", ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"zh", "zh", "", "ㄓ", ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL)},
{"zha", "zh", "a", "ㄓㄚ", ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_A)},
{"zhai", "zh", "ai", "ㄓㄞ", ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_AI)},
{"zhan", "zh", "an", "ㄓㄢ", ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_AN)},
{"zhang", "zh", "ang", "ㄓㄤ", ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_ANG)},
{"zhao", "zh", "ao", "ㄓㄠ", ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_AO)},
{"zhe", "zh", "e", "ㄓㄜ", ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_E)},
{"zhei", "zh", "ei", "ㄓㄟ", ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_EI)},
{"zhen", "zh", "en", "ㄓㄣ", ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_EN)},
{"zheng", "zh", "eng", "ㄓㄥ", ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_ENG)},
{"zhi", "zh", "i", "ㄓ", ChewingKey(CHEWING_ZH, CHEWING_I, CHEWING_ZERO_FINAL)},
{"zhong", "zh", "ong", "ㄓㄨㄥ", ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, PINYIN_ONG)},
{"zhou", "zh", "ou", "ㄓㄡ", ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"zhu", "zh", "u", "ㄓㄨ", ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_ZERO_FINAL)},
{"zhua", "zh", "ua", "ㄓㄨㄚ", ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_A)},
{"zhuai", "zh", "uai", "ㄓㄨㄞ", ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_AI)},
{"zhuan", "zh", "uan", "ㄓㄨㄢ", ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_AN)},
{"zhuang", "zh", "uang", "ㄓㄨㄤ", ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_ANG)},
{"zhui", "zh", "ui", "ㄓㄨㄟ", ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_EI)},
{"zhun", "zh", "un", "ㄓㄨㄣ", ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_EN)},
{"zhuo", "zh", "uo", "ㄓㄨㄛ", ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_O)},
{"zi", "z", "i", "ㄗ", ChewingKey(CHEWING_Z, CHEWING_I, CHEWING_ZERO_FINAL)},
{"zong", "z", "ong", "ㄗㄨㄥ", ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, PINYIN_ONG)},
{"zou", "z", "ou", "ㄗㄡ", ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_OU)},
{"zu", "z", "u", "ㄗㄨ", ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_ZERO_FINAL)},
{"zuan", "z", "uan", "ㄗㄨㄢ", ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_AN)},
{"zui", "z", "ui", "ㄗㄨㄟ", ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_EI)},
{"zun", "z", "un", "ㄗㄨㄣ", ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_EN)},
{"zuo", "z", "uo", "ㄗㄨㄛ", ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_O)}
};

const divided_table_item_t divided_table[] = {
{"bian", 182478, {"bi", "an"}, 100},
{"bie", 63919, {"bi", "e"}, 100},
{"dian", 179799, {"di", "an"}, 100},
{"jian", 435752, {"ji", "an"}, 200},
{"jiang", 139834, {"ji", "ang"}, 100},
{"jie", 294175, {"ji", "e"}, 100},
{"jue", 119987, {"ju", "e"}, 100},
{"kuai", 63367, {"ku", "ai"}, 100},
{"lian", 130021, {"li", "an"}, 100},
{"liang", 185438, {"li", "ang"}, 100},
{"liao", 39355, {"li", "ao"}, 100},
{"luan", 17609, {"lu", "an"}, 100},
{"qian", 195129, {"qi", "an"}, 100},
{"qie", 70219, {"qi", "e"}, 100},
{"shuan", 1114, {"shu", "an"}, 100},
{"tian", 185905, {"ti", "an"}, 100},
{"tuan", 17287, {"tu", "an"}, 100},
{"xian", 280991, {"xi", "an"}, 300},
{"yuan", 280423, {"yu", "an"}, 100},
{"zuan", 4016, {"zu", "an"}, 100}
};

const resplit_table_item_t resplit_table[] = {
{{"a", "nan"}, 0, {"an", "an"}, 100},
{{"an", "gang"}, 0, {"ang", "ang"}, 100},
{{"ba", "nan"}, 0, {"ban", "an"}, 100},
{{"ca", "nan"}, 0, {"can", "an"}, 100},
{{"chan", "gan"}, 0, {"chang", "an"}, 100},
{{"chan", "ge"}, 0, {"chang", "e"}, 100},
{{"che", "nai"}, 0, {"chen", "ai"}, 100},
{{"chen", "gan"}, 0, {"cheng", "an"}, 100},
{{"chu", "nan"}, 100, {"chun", "an"}, 100},
{{"dan", "gan"}, 0, {"dang", "an"}, 100},
{{"e", "nai"}, 0, {"en", "ai"}, 100},
{{"fa", "nan"}, 100, {"fan", "an"}, 100},
{{"fan", "gai"}, 0, {"fang", "ai"}, 100},
{{"fan", "gan"}, 100, {"fang", "an"}, 100},
{{"fan", "ge"}, 0, {"fang", "e"}, 100},
{{"ga", "nai"}, 0, {"gan", "ai"}, 100},
{{"ga", "nen"}, 0, {"gan", "en"}, 100},
{{"gan", "gao"}, 0, {"gang", "ao"}, 100},
{{"guan", "gan"}, 100, {"guang", "an"}, 100},
{{"hu", "nan"}, 100, {"hun", "an"}, 100},
{{"huan", "gan"}, 0, {"huang", "an"}, 100},
{{"ji", "ne"}, 0, {"jin", "e"}, 100},
{{"ji", "nou"}, 0, {"jin", "ou"}, 100},
{{"jia", "nai"}, 0, {"jian", "ai"}, 100},
{{"jia", "nan"}, 100, {"jian", "an"}, 100},
{{"jia", "ne"}, 0, {"jian", "e"}, 100},
{{"jia", "nou"}, 0, {"jian", "ou"}, 100},
{{"jian", "gan"}, 100, {"jiang", "an"}, 100},
{{"jin", "gai"}, 0, {"jing", "ai"}, 100},
{{"jin", "gan"}, 0, {"jing", "an"}, 100},
{{"jin", "ge"}, 0, {"jing", "e"}, 100},
{{"kuan", "gao"}, 0, {"kuang", "ao"}, 100},
{{"li", "nan"}, 100, {"lin", "an"}, 100},
{{"lia", "nai"}, 0, {"lian", "ai"}, 100},
{{"lia", "ne"}, 0, {"lian", "e"}, 100},
{{"lian", "gan"}, 0, {"liang", "an"}, 100},
{{"ma", "ne"}, 0, {"man", "e"}, 100},
{{"men", "gen"}, 0, {"meng", "en"}, 100},
{{"min", "gan"}, 100, {"ming", "an"}, 100},
{{"min", "ge"}, 100, {"ming", "e"}, 100},
{{"na", "nai"}, 0, {"nan", "ai"}, 100},
{{"na", "nan"}, 0, {"nan", "an"}, 200},
{{"na", "nao"}, 0, {"nan", "ao"}, 100},
{{"na", "nou"}, 0, {"nan", "ou"}, 100},
{{"nin", "gan"}, 0, {"ning", "an"}, 100},
{{"pa", "nan"}, 0, {"pan", "an"}, 100},
{{"pen", "gan"}, 0, {"peng", "an"}, 100},
{{"pin", "gan"}, 0, {"ping", "an"}, 100},
{{"qi", "nai"}, 0, {"qin", "ai"}, 100},
{{"qi", "nan"}, 0, {"qin", "an"}, 100},
{{"qia", "nan"}, 0, {"qian", "an"}, 200},
{{"qia", "ne"}, 0, {"qian", "e"}, 100},
{{"qin", "gai"}, 0, {"qing", "ai"}, 100},
{{"qin", "gan"}, 0, {"qing", "an"}, 100},
{{"re", "nai"}, 0, {"ren", "ai"}, 100},
{{"re", "nan"}, 0, {"ren", "an"}, 100},
{{"san", "gou"}, 0, {"sang", "ou"}, 100},
{{"shan", "gan"}, 100, {"shang", "an"}, 100},
{{"she", "nai"}, 0, {"shen", "ai"}, 100},
{{"she", "nao"}, 0, {"shen", "ao"}, 200},
{{"wa", "nan"}, 0, {"wan", "an"}, 200},
{{"wa", "ne"}, 0, {"wan", "e"}, 100},
{{"wa", "nou"}, 0, {"wan", "ou"}, 100},
{{"wen", "gan"}, 0, {"weng", "an"}, 100},
{{"xi", "nai"}, 200, {"xin", "ai"}, 100},
{{"xi", "nan"}, 100, {"xin", "an"}, 100},
{{"xia", "nai"}, 0, {"xian", "ai"}, 100},
{{"xia", "nan"}, 0, {"xian", "an"}, 100},
{{"xia", "ne"}, 0, {"xian", "e"}, 100},
{{"xian", "gai"}, 0, {"xiang", "ai"}, 100},
{{"xian", "gan"}, 200, {"xiang", "an"}, 100},
{{"xian", "ge"}, 100, {"xiang", "e"}, 100},
{{"xin", "gai"}, 0, {"xing", "ai"}, 100},
{{"xin", "gan"}, 200, {"xing", "an"}, 200},
{{"ya", "nan"}, 0, {"yan", "an"}, 200},
{{"yi", "nan"}, 300, {"yin", "an"}, 100},
{{"yi", "ne"}, 0, {"yin", "e"}, 100},
{{"zhan", "gai"}, 0, {"zhang", "ai"}, 100},
{{"zhe", "nai"}, 0, {"zhen", "ai"}, 200},
{{"zhe", "nan"}, 0, {"zhen", "an"}, 100},
{{"zhen", "gan"}, 100, {"zheng", "an"}, 100},
{{"zhua", "nan"}, 0, {"zhuan", "an"}, 100}
};

const gint chewing_key_table[CHEWING_NUMBER_OF_INITIALS *
                             CHEWING_NUMBER_OF_MIDDLES *
                             CHEWING_NUMBER_OF_FINALS] = {
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
1       /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
2       /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
3       /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
4       /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
5       /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
85      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
86      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
87      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
88      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
89      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
234     /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
252     /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
253     /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_I, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_ZERO_INITIAL, CHEWING_V, PINYIN_ING) */,
6       /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
7       /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
8       /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
9       /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
10      /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
11      /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
12      /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
13      /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
14      /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
21      /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
19      /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
20      /* ChewingKey(CHEWING_B, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
15      /* ChewingKey(CHEWING_B, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_I, CHEWING_AI) */,
16      /* ChewingKey(CHEWING_B, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_I, CHEWING_ANG) */,
17      /* ChewingKey(CHEWING_B, CHEWING_I, CHEWING_AO) */,
18      /* ChewingKey(CHEWING_B, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_I, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_I, PINYIN_ING) */,
22      /* ChewingKey(CHEWING_B, CHEWING_U, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_B, CHEWING_V, PINYIN_ING) */,
23      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
24      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
25      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
26      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
27      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
28      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
29      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
30      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
31      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
53      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
54      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
52      /* ChewingKey(CHEWING_C, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_I, PINYIN_ING) */,
55      /* ChewingKey(CHEWING_C, CHEWING_U, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_U, CHEWING_AI) */,
56      /* ChewingKey(CHEWING_C, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_U, INVALID_EA) */,
57      /* ChewingKey(CHEWING_C, CHEWING_U, CHEWING_EI) */,
58      /* ChewingKey(CHEWING_C, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_U, CHEWING_NG) */,
59      /* ChewingKey(CHEWING_C, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_C, CHEWING_V, PINYIN_ING) */,
32      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
33      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
34      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
35      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
36      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
37      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
38      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
39      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
40      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
42      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
43      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
41      /* ChewingKey(CHEWING_CH, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_I, PINYIN_ING) */,
44      /* ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_ZERO_FINAL) */,
45      /* ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_A) */,
46      /* ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_AI) */,
47      /* ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_AN) */,
48      /* ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_U, INVALID_EA) */,
49      /* ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_EI) */,
50      /* ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_NG) */,
51      /* ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_CH, CHEWING_V, PINYIN_ING) */,
60      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
61      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
62      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
63      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
64      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
65      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
66      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
67      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
68      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
69      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
78      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
79      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
75      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
76      /* ChewingKey(CHEWING_D, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
70      /* ChewingKey(CHEWING_D, CHEWING_I, CHEWING_ZERO_FINAL) */,
71      /* ChewingKey(CHEWING_D, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_I, CHEWING_AI) */,
72      /* ChewingKey(CHEWING_D, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_I, CHEWING_ANG) */,
73      /* ChewingKey(CHEWING_D, CHEWING_I, CHEWING_AO) */,
74      /* ChewingKey(CHEWING_D, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_I, PINYIN_ONG) */,
77      /* ChewingKey(CHEWING_D, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_I, PINYIN_ING) */,
80      /* ChewingKey(CHEWING_D, CHEWING_U, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_U, CHEWING_AI) */,
81      /* ChewingKey(CHEWING_D, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_U, INVALID_EA) */,
82      /* ChewingKey(CHEWING_D, CHEWING_U, CHEWING_EI) */,
83      /* ChewingKey(CHEWING_D, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_U, CHEWING_NG) */,
84      /* ChewingKey(CHEWING_D, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_D, CHEWING_V, PINYIN_ING) */,
90      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
91      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
92      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
93      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
94      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
95      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
96      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
97      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
98      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
99      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_I, PINYIN_ING) */,
100     /* ChewingKey(CHEWING_F, CHEWING_U, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_F, CHEWING_V, PINYIN_ING) */,
121     /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
122     /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
123     /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
124     /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
125     /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
126     /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
127     /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
128     /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
129     /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
130     /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
131     /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
132     /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_I, PINYIN_ING) */,
133     /* ChewingKey(CHEWING_H, CHEWING_U, CHEWING_ZERO_FINAL) */,
134     /* ChewingKey(CHEWING_H, CHEWING_U, CHEWING_A) */,
135     /* ChewingKey(CHEWING_H, CHEWING_U, CHEWING_AI) */,
136     /* ChewingKey(CHEWING_H, CHEWING_U, CHEWING_AN) */,
137     /* ChewingKey(CHEWING_H, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_U, INVALID_EA) */,
138     /* ChewingKey(CHEWING_H, CHEWING_U, CHEWING_EI) */,
139     /* ChewingKey(CHEWING_H, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_U, CHEWING_NG) */,
140     /* ChewingKey(CHEWING_H, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_H, CHEWING_V, PINYIN_ING) */,
101     /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
102     /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
103     /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
104     /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
105     /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
106     /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
107     /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
108     /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
109     /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
110     /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
111     /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
112     /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_I, PINYIN_ING) */,
113     /* ChewingKey(CHEWING_G, CHEWING_U, CHEWING_ZERO_FINAL) */,
114     /* ChewingKey(CHEWING_G, CHEWING_U, CHEWING_A) */,
115     /* ChewingKey(CHEWING_G, CHEWING_U, CHEWING_AI) */,
116     /* ChewingKey(CHEWING_G, CHEWING_U, CHEWING_AN) */,
117     /* ChewingKey(CHEWING_G, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_U, INVALID_EA) */,
118     /* ChewingKey(CHEWING_G, CHEWING_U, CHEWING_EI) */,
119     /* ChewingKey(CHEWING_G, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_U, CHEWING_NG) */,
120     /* ChewingKey(CHEWING_G, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_G, CHEWING_V, PINYIN_ING) */,
156     /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
157     /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
158     /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
159     /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
160     /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
161     /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
162     /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
163     /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
164     /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
165     /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
166     /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
167     /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_I, PINYIN_ING) */,
168     /* ChewingKey(CHEWING_K, CHEWING_U, CHEWING_ZERO_FINAL) */,
169     /* ChewingKey(CHEWING_K, CHEWING_U, CHEWING_A) */,
170     /* ChewingKey(CHEWING_K, CHEWING_U, CHEWING_AI) */,
171     /* ChewingKey(CHEWING_K, CHEWING_U, CHEWING_AN) */,
172     /* ChewingKey(CHEWING_K, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_U, INVALID_EA) */,
173     /* ChewingKey(CHEWING_K, CHEWING_U, CHEWING_EI) */,
174     /* ChewingKey(CHEWING_K, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_U, CHEWING_NG) */,
175     /* ChewingKey(CHEWING_K, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_K, CHEWING_V, PINYIN_ING) */,
141     /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
148     /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
149     /* ChewingKey(CHEWING_J, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
142     /* ChewingKey(CHEWING_J, CHEWING_I, CHEWING_ZERO_FINAL) */,
143     /* ChewingKey(CHEWING_J, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_I, CHEWING_AI) */,
144     /* ChewingKey(CHEWING_J, CHEWING_I, CHEWING_AN) */,
145     /* ChewingKey(CHEWING_J, CHEWING_I, CHEWING_ANG) */,
146     /* ChewingKey(CHEWING_J, CHEWING_I, CHEWING_AO) */,
147     /* ChewingKey(CHEWING_J, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_I, CHEWING_O) */,
150     /* ChewingKey(CHEWING_J, CHEWING_I, PINYIN_ONG) */,
151     /* ChewingKey(CHEWING_J, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_I, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_U, PINYIN_ING) */,
152     /* ChewingKey(CHEWING_J, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_V, CHEWING_AI) */,
153     /* ChewingKey(CHEWING_J, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_V, CHEWING_AO) */,
154     /* ChewingKey(CHEWING_J, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_V, CHEWING_EI) */,
155     /* ChewingKey(CHEWING_J, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_J, CHEWING_V, PINYIN_ING) */,
204     /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
205     /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
206     /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
207     /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
208     /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
209     /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
210     /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
211     /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
212     /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
213     /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
221     /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
222     /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
218     /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
219     /* ChewingKey(CHEWING_M, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
214     /* ChewingKey(CHEWING_M, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_I, CHEWING_AI) */,
215     /* ChewingKey(CHEWING_M, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_I, CHEWING_ANG) */,
216     /* ChewingKey(CHEWING_M, CHEWING_I, CHEWING_AO) */,
217     /* ChewingKey(CHEWING_M, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_I, PINYIN_ONG) */,
220     /* ChewingKey(CHEWING_M, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_I, PINYIN_ING) */,
223     /* ChewingKey(CHEWING_M, CHEWING_U, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_M, CHEWING_V, PINYIN_ING) */,
224     /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
225     /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
226     /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
227     /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
228     /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
229     /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
230     /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
231     /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
232     /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
233     /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
244     /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
245     /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
241     /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
242     /* ChewingKey(CHEWING_N, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
235     /* ChewingKey(CHEWING_N, CHEWING_I, CHEWING_ZERO_FINAL) */,
236     /* ChewingKey(CHEWING_N, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_I, CHEWING_AI) */,
237     /* ChewingKey(CHEWING_N, CHEWING_I, CHEWING_AN) */,
238     /* ChewingKey(CHEWING_N, CHEWING_I, CHEWING_ANG) */,
239     /* ChewingKey(CHEWING_N, CHEWING_I, CHEWING_AO) */,
240     /* ChewingKey(CHEWING_N, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_I, PINYIN_ONG) */,
243     /* ChewingKey(CHEWING_N, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_I, PINYIN_ING) */,
246     /* ChewingKey(CHEWING_N, CHEWING_U, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_U, CHEWING_AI) */,
247     /* ChewingKey(CHEWING_N, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_U, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_U, CHEWING_EI) */,
248     /* ChewingKey(CHEWING_N, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_U, CHEWING_NG) */,
249     /* ChewingKey(CHEWING_N, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_U, PINYIN_ING) */,
250     /* ChewingKey(CHEWING_N, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_V, CHEWING_AO) */,
251     /* ChewingKey(CHEWING_N, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_N, CHEWING_V, PINYIN_ING) */,
176     /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
177     /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
178     /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
179     /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
180     /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
181     /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
182     /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
183     /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
184     /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
185     /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
195     /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
196     /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
197     /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
192     /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
193     /* ChewingKey(CHEWING_L, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
186     /* ChewingKey(CHEWING_L, CHEWING_I, CHEWING_ZERO_FINAL) */,
187     /* ChewingKey(CHEWING_L, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_I, CHEWING_AI) */,
188     /* ChewingKey(CHEWING_L, CHEWING_I, CHEWING_AN) */,
189     /* ChewingKey(CHEWING_L, CHEWING_I, CHEWING_ANG) */,
190     /* ChewingKey(CHEWING_L, CHEWING_I, CHEWING_AO) */,
191     /* ChewingKey(CHEWING_L, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_I, PINYIN_ONG) */,
194     /* ChewingKey(CHEWING_L, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_I, PINYIN_ING) */,
198     /* ChewingKey(CHEWING_L, CHEWING_U, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_U, CHEWING_AI) */,
199     /* ChewingKey(CHEWING_L, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_U, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_U, CHEWING_EI) */,
200     /* ChewingKey(CHEWING_L, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_U, CHEWING_NG) */,
201     /* ChewingKey(CHEWING_L, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_U, PINYIN_ING) */,
202     /* ChewingKey(CHEWING_L, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_V, CHEWING_AO) */,
203     /* ChewingKey(CHEWING_L, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_L, CHEWING_V, PINYIN_ING) */,
287     /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
288     /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
289     /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
290     /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
291     /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
292     /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
293     /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
295     /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
296     /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
294     /* ChewingKey(CHEWING_R, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_I, PINYIN_ING) */,
297     /* ChewingKey(CHEWING_R, CHEWING_U, CHEWING_ZERO_FINAL) */,
298     /* ChewingKey(CHEWING_R, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_U, CHEWING_AI) */,
299     /* ChewingKey(CHEWING_R, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_U, INVALID_EA) */,
300     /* ChewingKey(CHEWING_R, CHEWING_U, CHEWING_EI) */,
301     /* ChewingKey(CHEWING_R, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_U, CHEWING_NG) */,
302     /* ChewingKey(CHEWING_R, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_R, CHEWING_V, PINYIN_ING) */,
254     /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
255     /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
256     /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
257     /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
258     /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
259     /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
260     /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
261     /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
262     /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
269     /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
270     /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
267     /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
268     /* ChewingKey(CHEWING_P, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
263     /* ChewingKey(CHEWING_P, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_I, CHEWING_AI) */,
264     /* ChewingKey(CHEWING_P, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_I, CHEWING_ANG) */,
265     /* ChewingKey(CHEWING_P, CHEWING_I, CHEWING_AO) */,
266     /* ChewingKey(CHEWING_P, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_I, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_I, PINYIN_ING) */,
271     /* ChewingKey(CHEWING_P, CHEWING_U, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_P, CHEWING_V, PINYIN_ING) */,
272     /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
279     /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
280     /* ChewingKey(CHEWING_Q, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
273     /* ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_ZERO_FINAL) */,
274     /* ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_AI) */,
275     /* ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_AN) */,
276     /* ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_ANG) */,
277     /* ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_AO) */,
278     /* ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_O) */,
281     /* ChewingKey(CHEWING_Q, CHEWING_I, PINYIN_ONG) */,
282     /* ChewingKey(CHEWING_Q, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_I, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_U, PINYIN_ING) */,
283     /* ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_AI) */,
284     /* ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_AO) */,
285     /* ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_EI) */,
286     /* ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_Q, CHEWING_V, PINYIN_ING) */,
303     /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
304     /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
305     /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
306     /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
307     /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
308     /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
309     /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
310     /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
311     /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
333     /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
334     /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
332     /* ChewingKey(CHEWING_S, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_I, PINYIN_ING) */,
335     /* ChewingKey(CHEWING_S, CHEWING_U, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_U, CHEWING_AI) */,
336     /* ChewingKey(CHEWING_S, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_U, INVALID_EA) */,
337     /* ChewingKey(CHEWING_S, CHEWING_U, CHEWING_EI) */,
338     /* ChewingKey(CHEWING_S, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_U, CHEWING_NG) */,
339     /* ChewingKey(CHEWING_S, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_S, CHEWING_V, PINYIN_ING) */,
312     /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
313     /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
314     /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
315     /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
316     /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
317     /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
318     /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
319     /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
320     /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
321     /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
323     /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
322     /* ChewingKey(CHEWING_SH, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_I, PINYIN_ING) */,
324     /* ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_ZERO_FINAL) */,
325     /* ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_A) */,
326     /* ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_AI) */,
327     /* ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_AN) */,
328     /* ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_U, INVALID_EA) */,
329     /* ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_EI) */,
330     /* ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_NG) */,
331     /* ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_SH, CHEWING_V, PINYIN_ING) */,
340     /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
341     /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
342     /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
343     /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
344     /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
345     /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
346     /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
347     /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
353     /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
354     /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
352     /* ChewingKey(CHEWING_T, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
348     /* ChewingKey(CHEWING_T, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_I, CHEWING_AI) */,
349     /* ChewingKey(CHEWING_T, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_I, CHEWING_ANG) */,
350     /* ChewingKey(CHEWING_T, CHEWING_I, CHEWING_AO) */,
351     /* ChewingKey(CHEWING_T, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_I, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_I, PINYIN_ING) */,
355     /* ChewingKey(CHEWING_T, CHEWING_U, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_U, CHEWING_AI) */,
356     /* ChewingKey(CHEWING_T, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_U, INVALID_EA) */,
357     /* ChewingKey(CHEWING_T, CHEWING_U, CHEWING_EI) */,
358     /* ChewingKey(CHEWING_T, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_U, CHEWING_NG) */,
359     /* ChewingKey(CHEWING_T, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_T, CHEWING_V, PINYIN_ING) */,
360     /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
367     /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, CHEWING_AI) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, CHEWING_ANG) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, CHEWING_AO) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, PINYIN_ONG) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_I, PINYIN_ING) */,
369     /* ChewingKey(PINYIN_W, CHEWING_U, CHEWING_ZERO_FINAL) */,
361     /* ChewingKey(PINYIN_W, CHEWING_U, CHEWING_A) */,
362     /* ChewingKey(PINYIN_W, CHEWING_U, CHEWING_AI) */,
363     /* ChewingKey(PINYIN_W, CHEWING_U, CHEWING_AN) */,
364     /* ChewingKey(PINYIN_W, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_U, INVALID_EA) */,
365     /* ChewingKey(PINYIN_W, CHEWING_U, CHEWING_EI) */,
366     /* ChewingKey(PINYIN_W, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_U, CHEWING_NG) */,
368     /* ChewingKey(PINYIN_W, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(PINYIN_W, CHEWING_V, PINYIN_ING) */,
370     /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
377     /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
378     /* ChewingKey(CHEWING_X, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
371     /* ChewingKey(CHEWING_X, CHEWING_I, CHEWING_ZERO_FINAL) */,
372     /* ChewingKey(CHEWING_X, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_I, CHEWING_AI) */,
373     /* ChewingKey(CHEWING_X, CHEWING_I, CHEWING_AN) */,
374     /* ChewingKey(CHEWING_X, CHEWING_I, CHEWING_ANG) */,
375     /* ChewingKey(CHEWING_X, CHEWING_I, CHEWING_AO) */,
376     /* ChewingKey(CHEWING_X, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_I, CHEWING_O) */,
379     /* ChewingKey(CHEWING_X, CHEWING_I, PINYIN_ONG) */,
380     /* ChewingKey(CHEWING_X, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_I, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_U, PINYIN_ING) */,
381     /* ChewingKey(CHEWING_X, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_V, CHEWING_AI) */,
382     /* ChewingKey(CHEWING_X, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_V, CHEWING_AO) */,
383     /* ChewingKey(CHEWING_X, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_V, CHEWING_EI) */,
384     /* ChewingKey(CHEWING_X, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_X, CHEWING_V, PINYIN_ING) */,
385     /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
393     /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
394     /* ChewingKey(PINYIN_Y, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
392     /* ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_ZERO_FINAL) */,
386     /* ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_A) */,
387     /* ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_AI) */,
388     /* ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_AN) */,
389     /* ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_ANG) */,
390     /* ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_AO) */,
391     /* ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_NG) */,
395     /* ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_O) */,
396     /* ChewingKey(PINYIN_Y, CHEWING_I, PINYIN_ONG) */,
397     /* ChewingKey(PINYIN_Y, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_I, PINYIN_ING) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, CHEWING_AI) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, INVALID_EA) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, CHEWING_EI) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, CHEWING_NG) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_U, PINYIN_ING) */,
398     /* ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_AI) */,
399     /* ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_AO) */,
400     /* ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_EI) */,
401     /* ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(PINYIN_Y, CHEWING_V, PINYIN_ING) */,
402     /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
403     /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
404     /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
405     /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
406     /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
407     /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
408     /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
409     /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
410     /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
411     /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
434     /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
435     /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
433     /* ChewingKey(CHEWING_Z, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_I, PINYIN_ING) */,
436     /* ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_AI) */,
437     /* ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_U, INVALID_EA) */,
438     /* ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_EI) */,
439     /* ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_NG) */,
440     /* ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_Z, CHEWING_V, PINYIN_ING) */,
412     /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_ZERO_FINAL) */,
413     /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_A) */,
414     /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_AI) */,
415     /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_AN) */,
416     /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_ANG) */,
417     /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_AO) */,
418     /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, INVALID_EA) */,
419     /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_EI) */,
420     /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_EN) */,
421     /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_O) */,
423     /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, PINYIN_ONG) */,
424     /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_ZERO_MIDDLE, PINYIN_ING) */,
422     /* ChewingKey(CHEWING_ZH, CHEWING_I, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_I, PINYIN_ING) */,
425     /* ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_ZERO_FINAL) */,
426     /* ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_A) */,
427     /* ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_AI) */,
428     /* ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_AN) */,
429     /* ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_U, INVALID_EA) */,
430     /* ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_EI) */,
431     /* ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_NG) */,
432     /* ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_U, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_U, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_U, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_U, PINYIN_ING) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, CHEWING_ZERO_FINAL) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, CHEWING_A) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, CHEWING_AI) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, CHEWING_AN) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, CHEWING_ANG) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, CHEWING_AO) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, CHEWING_E) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, INVALID_EA) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, CHEWING_EI) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, CHEWING_EN) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, CHEWING_ENG) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, CHEWING_ER) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, CHEWING_NG) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, CHEWING_O) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, PINYIN_ONG) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, CHEWING_OU) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, PINYIN_IN) */,
-1      /* ChewingKey(CHEWING_ZH, CHEWING_V, PINYIN_ING) */
};

};

#endif
